__all__ = ['myPod', 'genericIndent', 'delete', 'pairDeleteFirst',
           'pairDeleteSecond', 'mapUfunc', 'foreach',
           'intrinsicTypeCheck']

def genericIndent(src, globalDict):
    py_code = compile(src, '', 'exec')
    exec py_code
    block = ''
    indent = globalDict.get('indent')
    for line in code.splitlines():
        block += indent(tab, line)+'\n'
    return block

def delete(src, globalDict):
    py_code = compile(src, '', 'exec')
    exec py_code
    return code

def intrinsicTypeCheck(src, globalDict):
    '''A code insertion pod.

    @description
    This pod inserts the provided code block in `code` if the provided
    string `theType` is NOT a Fortran builtin intrinsic data type.

    @param[in] theType
    A string to compare if its a built-in Fortran intrsinic type.  For
    example, `integer`.

    @param[in] code
    The code block to insert if the `theType` is NOT a Fortran instrinsic
    type.

    '''
    py_code = compile(src, '', 'exec')
    exec py_code
    if not globalDict.get('FortranBlock').intrinsicType(theType):
        return code
    else:
        return ''

def pairDeleteFirst(src, globalDict):
    '''This pod is a work around for the moment since only one pod is allowed
    within a block due to pod naming restrictions.  Therefore, this serves as
    a wrapper for instrinicTypeCheck.

    '''
    return intrinsicTypeCheck(src, globalDict)

def pairDeleteSecond(src, globalDict):
    '''This pod is a work around for the moment since only one pod is allowed
    within a block due to pod naming restrictions.  Therefore, this serves as
    a wrapper for instrinicTypeCheck.

    '''
    return intrinsicTypeCheck(src, globalDict)

def mapUfunc(src, globalDict):
    code = compile(src, '', 'exec')
    exec code
    f90Block = globalDict.get('FortranBlock')
    tab = globalDict.get('tab')
    indent = globalDict.get('indent')
    f90Code = '''
    Object :: value, res

    type (InitError) :: err
    type (InsertionError) :: ierr

    interface
       function fnc(value)  result(res)
         InterfaceUseStatement
         Object :: value
         Object :: res
       end function fnc
    end interface

    ! Reinitialize the list
    err = init(other)

    ! Do the mapping
    if (associated(this%data)) then
       do while (iterate(this, value))
          res = fnc(value)
          ierr = push(other, res)
       enddo
    endif

'''
    f90Code = f90Code.replace('Type', 'type ('+thisType+')')
    if (f90Block.intrinsicType(obj)):
#        f90Code = f90Code.replace('InterfaceUseStatement', '')
        f90Code = f90Code.replace('InterfaceUseStatement', 'autouse\n end autouse\n')
        f90Code = f90Code.replace('Object', obj)
    else:
#        f90Code = f90Code.replace('InterfaceUseStatement', 'use class_'+obj)
        f90Code = f90Code.replace('InterfaceUseStatement', 'autouse\n end autouse\n')
        f90Code = f90Code.replace('Object', 'type ('+obj+')')
    return f90Code
    

def foreach(src, globalDict):
    """A simple looping pod function.

    @description
    The foreach function can be used in a pod or in a prepod.

    @param[in] src
    The src argument is a code listing from the prepod block. It should
    contain a variable called repl, a list called args, and a string
    called f90Code which contains the code upon which to do the substitutions
    and return as the new Fortran code.

    @param[in] globalDict
    This is the global dictionary for the present state of blockit.

    @warning
    If used in a pod, the foreach cannot have any function, subroutine,
    type, etc. declarations contained within. The resultant code is not
    parsed by the blockit framework and thus anything contained therein
    is not registered. It can still be used to replicate chunks of code.

    It is recommended that this only be used as a prepod.

    @note
    There is currently nothing in this function which makes it truly
    Fortran specific, it simply calls the code f90Code.

    @note
    I just defined the SrcCode variable in addition to f90Code so that
    it is less Fortran-specific.

    """

    repl = None
    args = None
    f90Code = None
    SrcCode = None
    code = compile(src, '', 'exec')
    exec code

    podcode = None
    if f90Code is not None: podcode = f90Code
    if SrcCode is not None: podcode = SrcCode
        
    block = ''
    for arg in args:
        subs = (len(repl) == 1) and [(arg, repl[0])] or zip(arg, repl) 
        tmp = podcode
        for sub,token in subs:
            tmp = tmp.replace('$'+token+'$', sub)
        block += tmp

    return block


