module Parameters
  
  use, intrinsic :: ISO_C_BINDING
  use class_Exception, only : DeallocError

  implicit none
  
  public :: MAX_DESC_LEN
  public :: dp
  public :: sp
  public :: short
  public :: long
  public :: delete
  public :: operator(==)
  public :: operator(<=)
  public :: operator(>=)
  public :: operator(<)
  public :: operator(>)
  private :: DeallocError
  private :: C_PTR_delete
  private :: C_PTR_equalTo
  private :: C_PTR_greaterThanEqualTo
  private :: C_PTR_lessThanEqualTo
  private :: C_PTR_greaterThan
  private :: C_PTR_lessThan

  integer, parameter :: MAX_DESC_LEN = 64
  integer, parameter :: dp = kind(1.d0)
  integer, parameter :: sp = kind(1.0)
  integer, parameter :: short = SELECTED_INT_KIND(9)
  integer, parameter :: long = SELECTED_INT_KIND(18)

  interface delete
     module procedure C_PTR_delete
  end interface

  interface operator(==)
     module procedure C_PTR_equalTo
  end interface

  interface operator(>=)
     module procedure C_PTR_greaterThanEqualTo
  end interface

  interface operator(<=)
     module procedure C_PTR_lessThanEqualTo
  end interface

  interface operator(>)
     module procedure C_PTR_greaterThan
  end interface

  interface operator(<)
     module procedure C_PTR_lessThan
  end interface

contains
  !==========================================================================
  function C_PTR_delete( this )  result( err )
    !.......................................................................
    ! Fake destructor for c_ptr
    !.......................................................................
    implicit none
    type (c_ptr), intent(inout) :: this
    type (DeallocError)         :: err

    err % error = 0
    this = C_NULL_PTR
  end function C_PTR_delete

  !.............................................................................
  ! Overloaded operators for c_ptr
  !.............................................................................
  !.............................................................................
  function C_PTR_lessThanEqualTo( this, other )  result( b )
    !...........................................................................
    ! <= Comparison of two c_ptr's
    !...........................................................................
    implicit none
    type (c_ptr), intent(in) :: this
    type (c_ptr), intent(in) :: other
    integer(long)            :: int_long

    logical :: b

    b = TRANSFER( this, int_long ) <= TRANSFER( other, int_long )
    
  end function C_PTR_lessThanEqualTo

  !................................................................................
  function C_PTR_greaterThanEqualTo( this, other ) result( b )
    !................................................................................
    ! >= Comparison of two c_ptr's
    !................................................................................
    implicit none
    type (c_ptr), intent(in) :: this
    type (c_ptr), intent(in) :: other
    integer(long)            :: int_long

    logical :: b

    b = TRANSFER( this, int_long ) >= TRANSFER( other, int_long )

  end function C_PTR_greaterThanEqualTo

  !................................................................................
  function C_PTR_equalTo( this, other ) result( b )
    !................................................................................
    ! >= Comparison of two c_ptr's
    !................................................................................
    implicit none
    type (c_ptr), intent(in) :: this
    type (c_ptr), intent(in) :: other
    integer(long)            :: int_long

    logical :: b

    b = TRANSFER( this, int_long ) == TRANSFER( other, int_long )

  end function C_PTR_equalTo

  !................................................................................
  function C_PTR_greaterThan( this, other ) result( b )
    !.............................................................
    ! > Comparison of two c_ptr's
    !.............................................................
    implicit none
    type (c_ptr), intent(in) :: this
    type (c_ptr), intent(in) :: other
    integer(long)            :: int_long

    logical :: b

    b = TRANSFER( this, int_long ) > TRANSFER( other, int_long )

  end function C_PTR_greaterThan

  !................................................................................
  function C_PTR_lessThan( this, other ) result( b )
    !.................................................................
    ! < Comparison of two c_ptr's
    !.................................................................
    implicit none
    type (c_ptr), intent(in) :: this
    type (c_ptr), intent(in) :: other
    integer(long)            :: int_long

    logical :: b

    b = TRANSFER( this, int_long ) < TRANSFER( other, int_long )

  end function C_PTR_lessThan

  !..........................................................................
  
end module Parameters
