#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
# Copyright (c) 2009,2010,2011 Michael List, michael.list@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
__all__ = ['FortranBlock',
           'AFile', 'PyF95Directives',
           'Template', 'Specialization',
           'Type',
           'Subroutine',
           'Module',
           'Function',
           'Program',
           'AutoInterface', 'AutoPrivate',
           'AutoUse',
           'AutoDelete', 'GarbageCollect',
           'Enumeration',
           'Foreach',
           'Instantiate',
           'Interface',
           'InterfaceA',
           'TestRunner', 'ParallelTestRunner',
           'createSymMap',
           'Pod', 'Prepod', 'isPod', 'isPrepod',
           'isFunction', 'isModule', 'isAType', 'isAFile', 'isInterface',
           'isProgram', 'isSubroutine', 'isTemplate', 'isAutoInterface',
           'isAutoPrivate', 'isTemplated', 'isAutoDelete', 'isRoot',
           'hasAsName',
           'isPyF95Directives',
           'hasTempParent',
           'parseTemplateSpecString',
           'mergeTemplateAndObjects',
           'templateSpecToTuple',
           'izip_longest',
           'globalSymbolTable',
           'globalSymbolMap' ]

from blockit.funcs import *
from blockit.blocks import Block
from blockit.sets import ASet
from blockit.scanners import LineScanner
from blockit.tables import SymbolTable
from blockit.parsers import pyp as pp
from parsers import FortranBlockParser
from grammars import *
try:
    from itertools import izip_longest
except ImportError:
    from itertools import chain
    from itertools import repeat
    from itertools import izip
#    def chain(*iterables):
#        # chain('ABC', 'DEF') --> A B C D E F
#        for it in iterables:
#            for element in it:
#                yield element
#
#    def repeat(object, times=None):
#        # repeat(10, 3) --> 10 10 10
#        if times is None:
#            while True:
#                yield object
#        else:
#            for i in xrange(times):
#                yield object

    def izip_longest(*args, **kwds):
        # izip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
        fillvalue = kwds.get('fillvalue')
        def sentinel(counter = ([fillvalue]*(len(args)-1)).pop):
            yield counter()         # yields the fillvalue, or raises IndexError
        fillers = repeat(fillvalue)
        iters = [chain(it, sentinel(), fillers) for it in args]
        try:
            for tup in izip(*iters):
                yield tup
        except IndexError:
            pass
#from blockit.Fortran.pod import isPrepod

import os
import re
import hashlib
import copy
import pickle
from os.path import basename as baseFileName

globalSymbolTable = SymbolTable( '|' )
globalSymbolMap = SymbolTable()
globalDependencyTable = SymbolTable()

#................................................................................
# Replacement functions for stuff not in python 2.5
#................................................................................
# def izip_longest( l1, l2, default ):
#     diff = len( l1 ) - len( l2 )
#     _l = zip( l1, l2 )
#     if diff > 0:
#         _l.extend( [( x, default ) for x in l1[ (len( l1 ) - diff): ]] )
#     elif diff < 0:
#         diff = - diff
#         _l.extend( [( default, x ) for x in l2[ (len( l2 ) - diff): ]] )

#     return _l

#................................................................................
# Filter functions for blocks
#................................................................................
isPod = lambda (x,v): isinstance(v, Pod)
isPrepod = lambda (x,v): isinstance(v, Prepod)
isForeach = lambda (x,v): isinstance(v, Foreach)
isFunction = lambda (x,v): isinstance(v, Function)
isModule = lambda (x,v): isinstance(v, Module)
isAType = lambda (x,v): isinstance(v, Type)
isAFile = lambda (x,v): isinstance(v, AFile)
isInterface = lambda (x,v): isinstance(v, Interface)
isProgram = lambda (x,v): isinstance(v, Program)
isSubroutine = lambda (x,v): isinstance(v, Subroutine)
isTemplate = lambda (x,v): isinstance(v, Template)
isAutoInterface = lambda (x,v): isinstance(v, AutoInterface)
isAutoPrivate = lambda (x,v): isinstance(v, AutoPrivate)
isAutoDelete = lambda (x,v): isinstance(v, AutoDelete)
isTemplated = lambda (x,v): v.isTemplated()
isRoot = lambda (x,v): ((v.parent() is None) or (isinstance(v.parent(), AFile)))
isSpecialization = lambda (x,v): isinstance(v, Specialization)
isPyF95Directives = lambda (x,v) : isinstance(v, PyF95Directives)
hasAsName = lambda (x,v): hasattr(v, 'asName') and v.asName is not None
hasTempParent = lambda (x,v): isinstance(v.parent(), Template)

#================================================================================
# Global functions for template handling
#================================================================================
def parseTemplateSpecString(s):
    '''Scan the template string and parse out the defined templates.  Returns
    a list of tuples with each type and its template specification.  Uses the
    TEMPLATE_DECL grammar.

    Examples:
    ---------
    >>> parseTemplateSpecString('DLList<List<integer>, Dict<String, List<integer>>>')
    >>> [('DLList', 'List<integer>,Dict<String,List<integer>>')]

    >>> parseTemplateSpecString('List<integer>,Dict<String,List<integer>>')
    >>> [('List', 'integer'), ('Dict', 'String,List<integer>')]

    '''
    return [(x[0], ''.join(x[1][1:-1]))
            for x in TEMPLATE_DECL.searchString(s)]

#================================================================================
def mergeTemplateAndObjects( cls, objs ):
    '''Merge a class pairs list with an objects tuple and return a substitution
    pair list.  If there is no object to pair with a class, use the default
    specified. For example:

    >>> classes = [('object', None), ('element', 'integer')]
    >>> objects = ('String', '*')
    >>> mergeTemplateAndObjects(classes, objects)
    [('object', 'String'), ('element', 'integer')]

    '''
    choose = lambda x,y: x!='*' and x or y
    if len( objs ) < len( cls ):
        if any( [ (x is None)  for x,y in cls[ ( len( objs ) - 1 ): ] ] ):
            msg = '''Insufficient template parameter specification and no default values for
missing parameters.'''
            raise Exception( msg )
    elif len( objs ) > len( cls ):
        msg = '''More template parameters specified than are available.'''
        raise Exception( msg )

    return [ ( c[ 0 ], ( choose( o, str( c[ 1 ] ) ) ) )
            for c,o in izip_longest( cls, objs, fillvalue='*' )]

#================================================================================
def templateSpecToTuple( string_ ):
    '''Explode a template string into its parts and return a tuple of the
    components.

    Example:
    --------
    >>> templateSpecToTuple('integer, AList<integer,real>')
    ('integer','AList<integer,real>')

    Input
    -----
    string_ : template spec string

    '''
    GRAMMAR = pp.delimitedList(pp.Combine(VALID_TEMPLATE_PARAM))
    # GRAMMAR = pp.delimitedList(VALID_TEMPLATE_PARAM, combine=True)
    return tuple(GRAMMAR.parseString(string_).asList())

#================================================================================
def createSymMap( table ):
    '''Loop through all the files blocks and resolve the symbols in the
    dependency lists.  Return a dictionary of names with mappings

    '''

    files = table.filter( isAFile )
    symMap = {}

    for blk in files.values():
        symMap.update( map( lambda x: ( x, blk.symbolTuple( x ) ),
                            [ x for x in blk.dependencies() ] ) )

    return symMap

#=========================
# Derived Block classes
#=========================
class FortranBlock(Block):
    #................................................................................
    # Intrinsic types list
    #................................................................................
    intrinsicList = ['integer', 'double', 'character', 'logical', 'complex',
                     'real']

    #................................................................................
    # pyparsing grammars for internal dependencies within a block
    #................................................................................
    BLOCKTYPE = 'FortranBlock'
    SENTINEL = None
    VERSION = '1.3.1'
    #................................................................................

    def __init__(self, blockType, parameters, childToken=None):
        '''Base FortranBlock class that all F90/95/2003 block types should inherit from.

        '''
        name = parameters[0]
        self._modLine = ''
        self._parameters = parameters
        self._objects = ASet()
        self._template = None
        self._use = ASet()
        self._types = SymbolTable()
        self._types['By Type'] = SymbolTable()
        self._types['By Symbol'] = SymbolTable()
        self._sigs = SymbolTable()
        self._version = FortranBlock.VERSION
        self._name_length = 31

        super(FortranBlock, self).__init__(blockType, name, childToken=childToken)

        #............................................................
        # Set parse actions for the dependency grammars
        #............................................................
        self.DEP_GRAMMAR = \
            ( TEMPLATE_TYPE.copy() ).setParseAction( self.__addType ) | \
            ( TEMPLATE_SUB.copy() ).setParseAction( self.__addSub ) | \
            ( TEMPLATE_FUNC.copy() ).setParseAction( self.__addFunc ) | \
            ( TEMPLATE_MOD_PROCEDURE_DECL.copy() ).setParseAction( self.__addProc ) | \
            ( INTRINSIC_TYPE.copy() ).setParseAction( self.__addIntrinsic ) | \
            ( USE_DECL.copy() ).setParseAction( self.__addUse )
        self.SHOW_GRAMMARS = \
            ( TEMPLATE_DECL('name') ).setParseAction( self.__mangleShow ) | \
            ( self.CHILDTOKEN.copy() ).setParseAction( self.__spawnChildShow )

    def getSymMap( self, table ):
        '''Get the symbol map for the block.

        '''
        return globalSymbolMap

    def version( self ):
        '''Return the block version number.

        '''
        return getattr( self, '_version', FortranBlock.VERSION )

    def contains( self, sym, skipList=None ):
        '''Search to see if the block has a symbol.

        '''
        if not self.isTemplated():
            if sym in self._children:  return True
            if skipList is not None:
                _children = [ x
                              for x in self._children.filter( Or( isModule,
                                                                  Or( isTemplate,
                                                                      isPrepod ) ) ).values()
                              if x not in skipList ]
            else:
                _children = self._children.filter( Or( isModule,
                                                       Or( isTemplate,
                                                           isPrepod ) ) ).values()
            for child in _children:
                if child.contains( sym, skipList ): return True
        return False

    def addLine( self, line, n=None ):
        '''Method to add a text line to a block code listing.  If the optional
        line number parameter is given, then the line is inserted AFTER the
        given line number.

        Input
        -----
        line : text line to add to code listing
        n : line number in listing to insert after.  Negative numbers insert
            from the end of the code listing.

        '''
        #......................................................................
        # Look for dependencies and propagate up the tree only if this is not
        # a templated block.  The reason is that templated blocks are
        # resolved in a second phase since they contain unresolved
        # declarations.  So, first any blocks that aren't templated may
        # declare an insantiation of a templated type.  These instantiations
        # are then propogated to the templated block.  After that, the block
        # can resolve all symbols and all instantiations can then be
        # resolved.  Dependencies for templated blocks are then known.
        #......................................................................
        self._modLine = line
        if not line.lstrip().startswith( '!' ) and not self.isTemplated():
            ans = self.DEP_GRAMMAR.searchString( line )

        self._addLine( n )

    def _addLine( self, n ):
        '''Add the internal _modLine to the code listing.

        '''
        if n is not None:
            self._code.insert( n, self._modLine )
        else:
            self._code.append( self._modLine )

    def __addType( self, s, l, t ):
        '''Add a scanned type declaration from a given pyparsed line.

        '''
        name = pack( t.name )
        _types = self._types
        self.addDeps( name, propagate=False )
        self._modLine = s.replace( t.name, name )
        if not _types[ 'By Type' ].has_key( name ):
            _types[ 'By Type' ][ name ] = []
        _types[ 'By Type' ][ name ].extend( t.symbolList )
        _types[ 'By Symbol' ].update( izip_longest( t.symbolList,
                                                    ( name,),
                                                    fillvalue=name ) )
    def __addIntrinsic( self, s, l, t ):
        '''Add a scanned instrinsic type declaration from a given pyparsed line.

        '''
        name = pack( t.name )
        if not self._types[ 'By Type' ].has_key( name ):
            self._types[ 'By Type' ][ name ] = []
        self._types[ 'By Type' ][ name ].extend( t.symbolList )
        self._types[ 'By Symbol' ].update( izip_longest( t.symbolList,
                                                        ( name, ),
                                                        fillvalue=name ) )

    def __addFunc( self, s, l, t ):
        '''Add a templated function call from a given pyparsed line.

        '''
        name = pack( t.name )
        self.addDeps( name, propagate=False )
        if t.args: self._sigs[ name ] = (t.args).asList()
        self._modLine = self._modLine.replace( t.name, name )

    def __addProc( self, s, l, t ):
        '''Add a procedure declaration from a given pyparsed line.

        '''
        name = pack( t.name )
        self.addDeps( name, propagate=False )
        self._modLine = self._modLine.replace( t.name, name )

    def __addSub( self, s, l, t ):
        '''Add a templated subroutine call from a given pyparsed line.

        '''
        name = pack( t.name )
        self.addDeps( name, propagate=False )
        if t.args: self._sigs[ name ] = t.args.asList()
        self._modLine = self._modLine.replace( t.name, name )

    def __addUse( self, s, l, t ):
        '''Add a scanned `use` declaration from a given pyparsed line.  Add to
        dependencies if its a templated use declaration.

        '''
        name = pack( t.name )
        self._use.add( name )
        if TEMPLATE_TOKENS[ 0 ] in name:
            self.addDeps( name, propagate=False )
        self._modLine = self._modLine.replace( t.name, name )

    def __addImport( self, s, l, t ):
        '''Add a scanned `import` declaration from a given pyparsed line.

        '''
        name = pack( t.name )
        self._modLine = self._modLine.replace( t.name, name )

    def rootNotAFile(self):
        if self._parent and not isinstance( self._parent, AFile ):
            return self._parent.rootNotAFile()
        return self

    def getState(self):
        '''Return the internal _state variable.

        '''
        return copy.copy( self._state )

    def setState(self, subs):
        '''Set the internal state variable _state.  Can be used for anything.

        '''
        self._state = copy.copy( subs )

    def isTemplated(self):
        '''Return true if the internal _template attribute is not None.  If
        the block has a parent, then call the parent isTemplated() method.
        A block is templated if it exists within a template block

        '''
        if self._template: return True
        if self._parent:
            return self._parent.isTemplated()
        return False

    def template(self):
        '''Return the _template attribute if this block has one.  If it
        does not and has a parent, the call the parents template() method,
        thereby walking up the tree.

        '''
        if self._template:  return self._template
        if self._parent and not isinstance( self._parent, AFile ):
            return self._parent.template()

    def addObjects(self, obj):
        '''Add a list of template specs to the internal objects set if this is
        a template block.  If not, propagate up the tree to find the parent
        template block.

        '''
        if self._template:
            self._objects.add(obj)
        elif self._parent:
            self._parent.addObjects(obj)
        else:
            raise Exception("Trying to add objects to a non-templated block!")

    def show(self, mangle=False):
        '''Show a realization of a block based on its current state.

        '''
        table = globalSymbolTable
        self._mangle = mangle
        block = ''
        for line in self._code:
            self._modLine = line
            ans = self.SHOW_GRAMMARS.searchString( line )
            block += self._modLine
        return block

    def __mangleShow( self, s, l, t ):
        '''Mangle the template tokens and set internal _modLine.

        '''
        _sym = ''.join( t.name )
        new_name = self.resolveSymbol( _sym )
        self._modLine = self._modLine.replace( _sym, new_name )

        # self.SHOW_GRAMMARS.searchString( self._modLine )
        # newName = self.resolveSymbol( t.name )
        # self._modLine = self._modLine.replace( t.name, newName )

    def __spawnChildShow( self, s, l, t ):
        '''Spawn the show() method of a child token and set internal _modLine

        '''
        sym = t.symbol
        _blk = self._children.get(sym, None)
        if _blk is None:
            print 'Child %s has disappeared from block %s'%(sym, self.name())
            self._modLine = ''
        else:
            self._modLine = _blk.show(self._mangle)

    def resolveInternal(self):
        '''Resolve the internal child dependencies of instantiated templated types for
        a given block so that the appear in the proper order internally within
        the block.

        '''
        #.....................................................................
        # 1. Get all children who are templates and instantiated, i.e. is a
        #    Template type but is not templated (brain melting).
        # 2. Create a full symbol table from these blocks.
        # 3. Resolve the order same as for files using topoSort.
        # 4. Reorder the lines in the file.
        # 5. Visit each child and goto 1.
        #.....................................................................
        loc = []
        childList = [self]
        inTable = SymbolTable()

        while len(childList) > 0:
            _blk = childList.pop(0)
            code = _blk.codeList()
            temps = _blk.children().filter(And(isTemplate,
                                               Not(isTemplated)))#.values()
            inTable.registerAll(temps)
            childList.extend(_blk.children().filter(Not(isTemplated)).values())

            if not temps: continue

            #....................................................................
            # Instantiated templates found.  Build diGraph and get ordering.  Use
            # insertion sort to reorganize blocks internally.
            #....................................................................
            diGraph = depGraph(temps.values(), inTable)
            order = topoSort(diGraph, prune=True)

            # The current code listing
            for t in order:
                loc.append((code.index(_blk.childString(t)), t))

            # Insertion sort... expensive, but list is generally small, so O(n^2)
            # is not too expensive
            while len(loc) > 0:
                n1, name = loc.pop(0)
                for i,pair in enumerate(loc):
                    n2, next = pair
                    # Swap code lines
                    if n1 > n2:
                        code[n1], code[n2] = code[n2], code[n1]
                        loc[0] = (n1, next)
                        break

            inTable.clear()

    @staticmethod
    def mangle(name, width=31, tempTokens=TEMPLATE_TOKENS):
        '''Generate a mangled name from a name and and a template string.

        Long names are contracted to fit within `width` argument size.

        '''
        def shorten( s, max_length ):
            s = '_'.join( [ x[ :-1 ] for x in s.split( '_' ) ] )
            if len( s ) > max_length:
                return shorten( s, max_length )
            return s

        tok1, tok2 = tempTokens
        i = name.find( tok1 )
        if i == -1: return name

        _name = name[ :i ]
        _s = name[ i: ] #_s = tok1 + s + tok2 # Encase in template characters
        temp = pack(_s).replace(tok1,'_').replace(tok2,'')
        temp = temp.replace(',','_').replace('(','_').replace(')','')
        temp = temp[1:]  # skip first _ character

        sz = width - len(_name) - 1
        if len(temp) > sz:
            if sz <= 0: raise Exception("Symbol width " + str(width) + " not adequate for unique name!")
            temp = shorten( temp, width - len(_name) - 1 )

        return '_'.join( [ _name, temp ] )

    def resolveSymbol( self, sym, mangle=True ):
        '''Resolve a templated symbol name.  This should only be used after
        the globalSymbolMap has been generated.  Otherwise, it will be
        expensive if called many times with symbols that have wildcards (*) or
        missing template parameters.  If the symbol is not already in the
        globalSymbolMap, then add it after resolving the symbol, i.e. cache it
        for future use.

        '''
        if sym not in globalSymbolMap:
            _sym, _sym_mangle = self.symbolTuple( sym )
            globalSymbolMap[ sym ] = ( _sym, _sym_mangle )
        else:
            _sym, _sym_mangle = globalSymbolMap[ sym ]

        if mangle: return _sym_mangle
        return _sym

    def __resolveSymbol( self, sym ):
        '''Resolve a symbol name, i.e. search for wildcards in a template
        string and substitute the wildcards with the appropriate default
        value as well as missing template parameters

        sym : symbol name to resolve

        '''
        _sym = parseTemplateSpecString(pack(sym))
        if _sym:
            # Resolve wildcards if necessary
            name, spec = _sym[0]
            #..............................................................
            # Hack:  Needed to avoid trying to resolve 'template <a,b,c>'
            # statements within Specialization blocks
            #..............................................................
            if name in RESERVED_KEYWORDS: return sym

            table = globalSymbolTable
            jc = table.joinCharacter()
            blk = table.filter(startsWith(name+jc)).values()
            if len( blk ) != 1:
                blk = filter( lambda x: x.isTemplated(), blk )
                if len( blk ) > 1:
                    print "Found (%d) possible symbols:"%(len(blk),)
                    for i,b in enumerate(blk):
                        error_msg = (i,b.name(),
                                     b.parent().name(),
                                     b.upTo(AFile).fileName() )
                        print "(%d) (%s) in block (%s) in file (%s) "% error_msg
                    raise Exception("Cannot resolve symbol %s"%(sym,))
                elif len( blk ) == 0:
                    return pack( sym )
#                     if sym in self._depIgnore:
#                         return pack( sym )
#                     else:
#                         print 'ignores: ', self._depIgnore
#                         raise Exception("Cannot resolve symbol %s"%(sym,))
            blk = blk[0]
            diff = len( blk.template() ) - len( templateSpecToTuple( spec ) )
            spec += ',*' * diff

            try:
                pairs = mergeTemplateAndObjects(blk.template(),
                                                templateSpecToTuple(spec))
            except Exception as inst:
                print inst
                msg = "When trying to resolve symbol (%s)"%(sym,)
                raise Exception(msg)

            #.....................................................................
            # Recursive call to resolve wildcards and missing template parameters
            #.....................................................................
            _pairs = []
            for x,y in pairs:
                y = self.__resolveSymbol( y )
                _pairs.append( y )
            #......................................
            spec = ','.join(_pairs)

            return name + TEMPLATE_TOKENS[ 0 ] + spec + TEMPLATE_TOKENS[ 1 ]

        return sym

    def symbolTuple( self, x ):
        '''Given a template symbol 'x' (or regular symbol), return a tuple with
        that contains the symbol with all wildcards and missing template
        parameters resolved and the mangled version of the symbol.

        '''
        sym = self.__resolveSymbol( x )
        return ( sym, self.mangle( sym, width=self._name_length ) )

    def name( self, tempTokens=TEMPLATE_TOKENS ):
        '''Return the name of the block.

        '''
        if self._state:
            temp = ','.join([obj for k,obj in self._state])
            name = self._name + tempTokens[0] + temp + tempTokens[1]
            return FortranBlock.mangle( name, width=self._name_length, tempTokens=tempTokens )

        return self._name

    def useDeclarations( self ):
        '''Return the set of use declarations.

        '''
        return self._use.copy()

    def __getstate__( self, _dd=None ):
        if _dd is None: _dd = self.__dict__.copy()
        _dd.pop('DEP_GRAMMAR')
        _dd.pop('SHOW_GRAMMARS')
        return super( FortranBlock, self ).__getstate__(_dd)

    def serialize( self, _dd=None ):
        '''Builds a dictionary tree of the block.

        '''
        return self.__getstate__(_dd)

    def __setstate__( self, tree ):
        '''Unpack the dictionary tree.

        '''
        g = globals()
        self.__init__( tree[ '_parameters' ] )
        self.__dict__.update( tree )
        _children = self._children
        self._children = SymbolTable()

        for node in _children:
            _blockType = node[ '_blockType' ]
            _params = node[ '_parameters' ]
            child = g[ _blockType ]( _params )
            child.__setstate__( node )
            child.setParent( self )
            self._children.register( child )

    def getSig( self, name ):
        '''Return the call signature for the given templated
        function/subroutine name that this block calls.

        '''
        _args = self._sigs.get( name, None )

        # Cross reference the internal symbols and return a lists of types
        # for the signature
        if _args:
            return [ self.findSymbol( x ) for x in _args ]

    def sentinel( self, line ):
        '''Overrided sentinel method. Calls super first.  If true, then looks
        for AutoDelete/Specialization blocks in scope and merges symbols and
        types found in these blocks with its own symbols and types.

        '''
        _ans = super( FortranBlock, self ).sentinel( line )
        if _ans:
            for x in self._children.filter( Or( isAutoDelete,
                                                isSpecialization ) ).values():
                self._types['By Symbol'].update( x.symbols )
                self._types['By Type'].update( x.types )
        return _ans

    def findSymbol( self, sym ):
        '''Look for a type symbol in the block.  If not found, check the
        enclosing block.

        '''
        val = self._types['By Symbol'].get( sym, None )
        if val is None and self._parent: return self._parent.findSymbol( sym )
        return ( val is None and 'character' or val )

    def setNameLength( self, name_length ):
        """Set the max number of characters for mangled templated names.

        Arguments:
        - `name_length` : integer number of max characters
        """
        self._name_length = name_length

    def getNameLength( self ):
        """Return the max number of characters for mangled templated names.

        """
        return self._name_length

    def addChild( self, child, token=None, before=False, regen=True ):
        '''Overrided addChild method which simply is needed to set the
        internal _name_length parameter to that of the parent.

        '''
        super( FortranBlock, self ).addChild( child, token, before, regen )
        child.setNameLength( self._name_length )

    @staticmethod
    def intrinsicType(s):
        '''Function returning a boolean whether any of the defined built-in types
        are found in the given string.

        '''
        x = s.strip().split('(')[0]

        return any([(x in b) for b in FortranBlock.intrinsicList])

    #..........................................................................
    # Properties
    #..........................................................................
    def __propTypes():
        doc = "SymbolTable of types declared within block"
        def fget( self ):
            return self._types[ 'By Type' ]
        return locals()

    def __propSymbols():
        doc = "SymbolTable of type symbols declared within block"
        def fget( self ):
            return self._types[ 'By Symbol' ]
        return locals()

    types = property( **__propTypes() )
    symbols = property( **__propSymbols() )

#................................................................................
class AFile( FortranBlock ):
    # No GRAMMAR for file block
    GRAMMAR = None
    SENTINEL = None
    BLOCKTYPE = 'AFile'

    HASHTOKEN_STRING =  '!PyF95.hash: '
    SYMBOLTOKEN_STRING = '!PyF95.symbol: '
    HASHGRAMMAR = pp.Literal(HASHTOKEN_STRING)
    SYMBOLGRAMMAR = pp.Literal(SYMBOLTOKEN_STRING)

    def __init__(self, parameters):
        '''A file block.

        '''
        # Transform the name (basename of the filename)
        self._fileName = parameters[0]
        self._sourceFileName = parameters[0]
        parameters[0] = baseFileName(self._fileName)
        self._hash = parameters[1]
        super(AFile, self).__init__(self.BLOCKTYPE, parameters)

    def fileName(self):
        return self._fileName

    def sourceFileName(self):
        '''This returns the original file name that was parsed.  This is
        needed because many of the global templates have their file names
        changed so that they are written within the source tree.  The library
        needs the original file name to be able to associate instantiated
        templates that may be in this file with its source.

        '''
        return self._sourceFileName

    def setFileName(self, name):
        self._fileName = name

    def addFileDep( self, name ):
        self._crossFileDeps.add( name )

#................................................................................
class EmptyFile( AFile ):
    # No GRAMMAR for file block
    GRAMMAR = None
    SENTINEL = None
    BLOCKTYPE = 'EmptyFile'

#................................................................................
class Template(FortranBlock):
    #....................................................
    # pyparsing grammar
    #....................................................
    BLOCKTYPE = 'Template'
    KEYWORD =  pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    TEMPLATE_PARAM_TYPE = pp.CaselessKeyword('class')('type')
    DEFAULT_ASSIGNMENT = pp.Suppress('=')
    TEMPLATE_DECL_COMB = pp.Combine(TEMPLATE_DECL, adjacent=False)
    VALIDNAME_WITH_PAREN_COMB = pp.Combine(VALIDNAME_WITH_PAREN, adjacent=False)
    TEMPLATE_ITEM = \
        pp.Group(
        TEMPLATE_PARAM_TYPE+\
        VALIDNAME('name')+\
        pp.Optional(
            DEFAULT_ASSIGNMENT+\
                (TEMPLATE_DECL_COMB|
                 VALIDNAME_WITH_PAREN_COMB|
                 DECIMAL_NUM|
                 NUM)('default')))
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        pp.Suppress(LEFT_TEMPTOK)+\
        pp.Group(pp.delimitedList(TEMPLATE_ITEM))('specs')+\
        pp.Suppress(RIGHT_TEMPTOK)
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        pairs = [(x.name, (x.default or None)) for x in tokens.specs]

        return Template([pairs])

    def __init__(self, parameters, name=None):
        '''A Template block.  Template blocks begin with the line:

        template < ... >

        where ... is a list of template parameters starting with the keyword
        class.  For example,

        template <class A, class B>
        .
        .
        .
        end template

        is the start of a template block with two internal template paramters
        A and B.

        Input
        -----
        parameters : the groups from the factory static method
        name : optional name to give the template

        '''

        # There is no "name" for a template, so used the id() function to
        # generate a unique one if this is a template block.
        template = parameters[0]
        if not name:
            parameters = ['template_'+str(id(self))]
        else:
            parameters = [name]

        super(Template, self).__init__(self.BLOCKTYPE, parameters)
        self._template = template
        self._name_length = 256

    def _createTemplatePairs(self, tokens):
        """Create template pairs from the pyparsed tokens list.

        Arguments:
        - `self`:
        - `tokens`: nested tokens list from pyparsing
        """

        if len(tokens) > 0:
            return [(x.name, x.default) for x in tokens[0]]
        return []

    def isTemplated(self):
        '''Return True if the code block is templated or within a templated
        code block.
        '''
        if self._template:  return True
        return False

    def sentinel(self, line):
        '''Overrided sentinel method to strip first and last lines from code
        block.

        '''
        if self.SENTINEL.searchString(line):
            self._code.pop(0)
            return True

    def template(self):
        '''Return the template specification string of the template.

        '''
        return self._template

    def isTemplated(self):
        '''Returns True if template block is templated, i.e. it has a template
        specification.  Instantiated template blocks are template blocks that
        do not have template specs so they will return false.

        '''
        if self._template: return True
        return False

    def addObjects(self, obj):
        '''Added a specification set for the template, e.g. if the template
        specification is <class Object, class Element>, then if you want
        Object to be 'integer' and Element to be 'real' obj would be in the
        argument list would be ('integer', 'real').

        obj : specification of a template, e.g. ('integer', 'real') for the
              template string <class Object, class Element>.

        '''
        self._objects.add(obj)

    def show(self, mangle=False, force=False):
        '''Overloaded show method for templates.  Only show template if it is
        a realized template, i.e. the _template attribute is None.  This can
        be overridden with the force parameter to show it anyway.

        '''
        if not self._template or force:
            return super(Template, self).show(mangle)
        return ''

    def hash(self):
        '''Overloaded hash method for this block if it does not already have a
        _hash internal value.  Uses the mangled version for the hash.  The
        hash value is:

        <the text hash>/<template parameters>

        '''
        hsh = getattr(self, '_hash', None)
        if hsh is None:
            self._hash = hashlib.sha256(self.show(True, True)).hexdigest()

        if self._state:
            return '%s/<%s>'%(self._hash,','.join([y for x,y in self._state]))
        else:
            return self._hash

    def name(self):
        '''Name of templates does not adhere to 31 character limit

        '''
        return super(Template, self).name()

    def instantiate( self, objects, cache ):
        '''This method only exists for realizations of templated blocks.  The
        setState() method is called internally with the provide list of
        template specifications which sets the template parameter states.
        Then this method creates a realization of a template block using those
        parameters and returns it.

        objects : a template parameter list, e.g ('*','integer','*')
                  for a three parameter template with default values used for
                  the first and last value or ('integer', 'real') for a two
                  parameter list.  The wildcards are resolved internally.

        cache : and instance of FortranCacheManager to check for instantiated
                templates

        '''
        global globalSymbolMap, subDict

        #.....................................................
        # Instantiate children of template block
        #.....................................................
        def _instantiate( tempBlk, blk ):
            #............................................................
            # Go through code and spawn children while parsing.  Use
            # addChild() AFTER parsing to propagate dependencies to the
            # parent.
            #............................................................
            global subDict

            totalLines = len( tempBlk._code )
            for lineNo,line in enumerate( tempBlk._code ):
                if lineNo == 0 or lineNo == totalLines-1:
                    # Note:  Trick adding leading spaces
                    line = line.replace(' '+tempBlk.baseName(), ' '+blk.name())

                isChild = tempBlk.CHILDTOKEN.searchString( line )
                if isChild:
                    name = isChild[0].symbol

                    child = tempBlk.children()[ name ]
                    child.setState( tempBlk.getState() )

                    _params = list( child._parameters ) # parent block parameters
                    _params[0] = child.name()           # name of new block

                    newBlk = type( child )( _params )
                    newBlk.setNameLength( child.getNameLength() )
                    newBlk = _instantiate( child, newBlk )
                    blk.addChild( newBlk,
                                  child.childString( child.baseName() ) )
                    line = '' # empty string
                else:
                    for k,v in subDict.items():
                        if line.lstrip().startswith( k ) and not \
                                blk.intrinsicType( v ):
                            line = line.replace( k, 'type ('+ v +')')
                        else:
                            line = line.replace( k, v )

                if line: blk.addLine( line )

            # blk._hash = tempBlk.hash()
            tempBlk.setState( None )
            blk.setNameLength( tempBlk.getNameLength() )
            return blk

        #.....................................................................
        # Setup the new template block (that really isn't templated, but just
        # a container for its children) and call the recursive _instantiate()
        # Note:  Must reset the state after instantiation to ensure there are
        #        no name conflicts when re-populating SymbolTables.
        #
        # Need to resolve object names if they contain wildcards.  This is
        # done by calling resolveSymbol() on each object in the list.
        #.....................................................................
        try:
            diff = len( self.template() ) - len( objects )
            _objects = list( copy.copy( objects ) )
            _objects.extend( [ '*' ] * diff )
            _objects = [self.resolveSymbol(o, mangle=False) for o in _objects]
        except Exception as inst:
            msg = inst.args
            print msg
            msg = '''Failed while trying to instantiate %s with parameters
            %s''' % ( self.name(), str( objects ) )
            raise Exception( msg )

        pairs = mergeTemplateAndObjects( self.template(), _objects )
        self.setState( pairs )
        subDict = {}
        subDict.update( self.getState() )

        #......................................................................
        # Check to see if there is a cached version of the template.
        # The dictionary must be gotten before setting the templates state so
        # that the hash is correct to get the dictionary.  Once you get
        # the dictionary, then set the state and use that hash as the key into
        # the dictionary.
        #......................................................................
        _inCache = False
        _blk = cache.fromCache( self )

        if _blk:
            print '-> Found cached instantiation...'
            _inCache = True
            _blk.fromCache = True
        else:
            _blk = type( self )([None], name=self.name())
            _blk.setNameLength( self.getNameLength() )
            _blk = _instantiate( self, _blk )
            _blk.fromCache = False

        self._parent.addChild( _blk, self.childString( self.baseName() ) )

        #...................................................................
        # Set the internal _hash value of the instantiated template to the
        # template it was generated from.
        #...................................................................
        self.setState( None )
        if not _inCache: cache.toCache( _blk )

        # globalSymbolMap.update( map( lambda x: ( x, _blk.symbolTuple( x ) ),
        #                              [ x for x in _blk.dependencies() ] ) )
        return _blk

#................................................................................
class Program(FortranBlock):
    #....................
    # pyparsing grammar
    #....................
    BLOCKTYPE = 'Program'
    KEYWORD = pp.CaselessKeyword(BLOCKTYPE)
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        VALIDNAME('name')
    #......................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return Program([tokens.name])

    def __init__(self, parameters):
        '''A F90/95/2003 program block.

        program foo
        .
        .
        .
        end program foo

        Input
        -----
        parameters : the groups from the factory static method

        '''
        name = parameters[0]
        super(Program, self).__init__(self.BLOCKTYPE, parameters)

    def getState(self):
        return "Stub:  Program getState"

#................................................................................
class Module(FortranBlock):
    #.....................................................
    # pyparsing grammar
    #.....................................................
    BLOCKTYPE = 'Module'
    KEYWORD = pp.CaselessKeyword(BLOCKTYPE)
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        VALIDNAME('name') +\
        (pp.StringEnd()|pp.Literal(';'))
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return Module([tokens.name])

    def __init__(self, parameters):
        '''A F90/95/2003 Module block:

        module foo
        .
        .
        .
        end module foo

        Input
        -----
        parameters : the groups from the factory static method

        '''
        super(Module, self).__init__(self.BLOCKTYPE, parameters)

#................................................................................
class Function(FortranBlock):
    #....................
    # pyparsing grammar
    #....................
    BLOCKTYPE = 'Function'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    RECURSIVE = pp.Suppress(pp.CaselessKeyword('recursive'))
    ELEMENTAL = pp.Suppress(pp.CaselessKeyword('elemental'))
    RESULT = \
        pp.Suppress(pp.CaselessKeyword('result')) +\
        pp.Suppress('(') +\
        VALIDNAME('resName') +\
        pp.Suppress(')')
    VALID_ASNAME = pp.Word(pp.alphas, pp.alphanums + '_./()=<>*+-')
    AS_CLAUSE = pp.Suppress('as')+VALID_ASNAME('asName')
    NO_MANGLE = pp.Word('no_mangle')('noMangle')
    GRAMMAR = \
        pp.StringStart() +\
        pp.Optional(RECURSIVE|ELEMENTAL) +\
        KEYWORD +\
        VALIDNAME('funcName') +\
        ARGLIST('args') +\
        pp.Optional(RESULT) +\
        pp.Optional(AS_CLAUSE)+\
        pp.Optional(NO_MANGLE)
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        asName = tokens.asName or None
        noMangle = tokens.noMangle or None
        resName = tokens.resName or None
        parameters = [tokens.funcName, tokens.args.asList(),
                      resName, asName, noMangle]
        return Function(parameters)

    def __init__(self, parameters):
        '''A F90/95/2003 subroutine block:

        function foo(x,y,z)  result(bar)
        .
        .
        .
        end function foo

        Input
        -----
        parameters : the groups from the factory static method

        '''
        self._argList, self._resName, self._asName = parameters[1:4]
        if parameters[4] is None:
            self._no_mangle = False
        else:
            self._no_mangle = True
        # self._asName = parameters[3]
        super(Function, self).__init__(self.BLOCKTYPE, parameters)

    def addLine(self, line, n=None):
        '''Overloaded addLine() method for a function block.  It strips out
        the 'as' and 'no_mangle' clauses prior to calling the base class
        addLine() method.

        '''
        if ' as ' in line:  line = line[0:line.index(' as ')]+os.linesep
        if ' no_mangle' in line:  line = line[0:line.index(' no_mangle')]+os.linesep
        super(Function, self).addLine(line)

    def name(self):
        '''Overloaded name method.  If "no_mangle" was specified, returns the
        original name with no template mangling.  Otherwise, it calls the base
        class name() method.

        '''
        if self._no_mangle: return self._name
        return super(Function, self).name()

#     def asName( self ):
#         return self._asName

#     def argList( self ):
#         return self._argList

    #......................................................................
    # Properties
    #......................................................................
    def __propAsName():
        doc = "The `as` attribute of the function"
        def fget( self ):
            return self._asName
        return locals()

    def __propArgList():
        doc = "The argument list of the function"
        def fget( self ):
            return self._argList
        return locals()

    asName = property( **__propAsName() )
    argList = property( **__propArgList() )

#................................................................................
class Subroutine(FortranBlock):
    #....................................................
    # pyparsing grammar
    #....................................................
    BLOCKTYPE = 'Subroutine'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    RECURSIVE = pp.Suppress(pp.CaselessKeyword('recursive'))
    ELEMENTAL = pp.Suppress(pp.CaselessKeyword('elemental'))
    VALID_ASNAME = pp.Word(pp.alphas, pp.alphanums+'_./()=<>+-*')
    AS_CLAUSE = pp.Suppress('as')+VALID_ASNAME('asName')
    NO_MANGLE = pp.Word('no_mangle')('noMangle')
    GRAMMAR = \
        pp.StringStart() +\
        pp.Optional(RECURSIVE|ELEMENTAL) +\
        KEYWORD +\
        VALIDNAME('subName') +\
        ARGLIST('args') +\
        pp.Optional(AS_CLAUSE) +\
        pp.Optional(NO_MANGLE)
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        asName = tokens.asName or None
        noMangle = tokens.noMangle or None
        parameters = [tokens.subName, tokens.args.asList(), asName, noMangle]
        return Subroutine(parameters)

    def __init__(self, parameters):
        '''A F90/95/2003 subroutine block:

        subroutine foo(x,y,z)
        .
        .
        .
        end subroutine foo

        Input
        -----
        parameters : the groups from the factory static method

        '''
        self._argList, self._asName = parameters[1:3]
        if parameters[3] is None:
            self._no_mangle = False
        else:
            self._no_mangle = True
        self._template = None
        super(Subroutine, self).__init__(self.BLOCKTYPE, parameters)

    def addLine(self, line, n=None):
        '''Overloaded addLine() method for a subroutine block.  It strips out
        the 'as' and 'no_mangle' clauses prior to calling the base class
        addLine() method.

        '''
        if ' as ' in line:  line = line[0:line.index(' as ')]+os.linesep
        if ' no_mangle' in line:  line = line[0:line.index(' no_mangle')]+os.linesep
        super(Subroutine, self).addLine(line)

    def name(self):
        '''Overloaded name method.  If "no_mangle" was specified, returns the
        original name with no template mangling.  Otherwise, it calls the base
        class name() method.

        '''
        if self._no_mangle: return self._name
        return super(Subroutine, self).name()

#     def asName(self):
#         return self._asName

    #......................................................................
    # Properties
    #......................................................................
    def __propAsName():
        doc = "The `as` attribute of the function"
        def fget( self ):
            return self._asName
        return locals()

    def __propArgList():
        doc = "The argument list of the function"
        def fget( self ):
            return self._argList
        return locals()

    asName = property( **__propAsName() )
    argList = property( **__propArgList() )

#................................................................................
class Interface(FortranBlock):
    #.......................................................
    # pyparsing grammar
    #.......................................................
    BLOCKTYPE='Interface'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    OPERATOR = \
        pp.Combine(
        pp.CaselessKeyword('operator')('type') +\
        pp.Literal('(') +\
        ( pp.Combine( pp.Literal('.') + pp.Word( pp.alphas ) + pp.Literal('.') ) |
          pp.Literal('<=') |
          pp.Literal('>=') |
          pp.Literal('==') |
          pp.Literal('<') |
          pp.Literal('>') )('op') +\
        pp.Literal(')'), adjacent=False)
    ASSIGNMENT = \
        pp.Combine(
        pp.CaselessKeyword('assignment')('type') +\
        pp.Literal('(') +\
        pp.Literal('=') +\
        pp.Literal(')'), adjacent=False)
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        (OPERATOR|ASSIGNMENT|pp.originalTextFor(TEMPLATE_DECL)|VALIDNAME)('name')
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        name = tokens.name or None
        parameters = [name]
        return Interface(parameters)

    def __init__(self, parameters):
        '''A F90/95/2003 interface block:

        interface foo
        .
        .
        .
        end interface

        Input
        -----
        parameters : the groups from the factory static method

        '''
        # Send blank line for name to create the proper sentinel
        super(Interface, self).__init__(self.BLOCKTYPE, [''])
        if parameters[0] is None:
            self._name = 'interface_'+str(id(self))
        else:
            self._name = pack( parameters[0] )

    def name(self):
        # No mangled names for interface names within templates
        return self._name

    def addLine( self, line, n=None ):
        if len( self._code ) == 0:
            self._code.append( line )
        else:
            super( Interface, self ).addLine( line, n )

#................................................................................
class InterfaceA(FortranBlock):
    #.....................................................
    # pyparsing grammar
    #.....................................................
    BLOCKTYPE = 'Interface'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    OPERATOR = \
        pp.Combine(
        pp.CaselessKeyword('operator')('type') +\
        pp.Literal('(') +\
        pp.Word(pp.alphas+'.', pp.alphanums+'_.<=>+/-')('op') +\
        pp.Literal(')'), adjacent=False)
    ASSIGNMENT = \
        pp.Combine(
        pp.CaselessKeyword('assignment')('type') +\
        pp.Literal('(') +\
        pp.Literal('=') +\
        pp.Literal(')'), adjacent=False)
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        (OPERATOR|ASSIGNMENT|VALIDNAME)('name')
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        name = tokens.name or None
        parameters = [name]
        return InterfaceA(parameters)

    def __init__(self, parameters):
        # Send blank line for name to create the proper sentinel
        super(Interface, self).__init__(self.BLOCKTYPE, [''])
        if parameters[0] is None:
            self._name = 'interface_'+str(id(self))
        else:
            self._name = parameters[0]

    def name(self):
        '''Overloaded name() method.  No name mangling for interface names.

        '''
        # No mangled names for interface names within templates
        return self._name

    def show(self, mangle=False):
        '''Overloaded show method.  For user defined interfaces within templates this
        finds the new names of the functions/subroutines due to name mangling
        and rebuilds the interface block to reflect the new names.

        '''
        parent = self._parent
        if parent is not None:
            funcsAndSubs = parent.children().filter(Or(isFunction,
                                                       isSubroutine))
        else:
            funcsAndSubs = None

        if funcsAndSubs is None: return ''.join(self._code)
        if len(funcsAndSubs) == 0: return ''.join(self._code)

        block = ''
        for line in self._code:
            ans = self.procedureNames.search(line)
            if ans:
                names = ans.group(1).split(',')
                newNames = []
                for name in names:
                    name = name.rstrip()
                    fnc_sub = funcsAndSubs.filter(startsWith(name))
                    if len(fnc_sub) != 1:
                        raise Exception("Ambiguous or unresolved name in "+\
                                        "Interface block: %s%s"%(str(fnc_sub),os.linesep)+\
                                        "Name: %s"%(name,))
                    fnc_sub = fnc_sub.values()[0]
                    line = line.replace(name, fnc_sub.name())
            block += line
        return block

#................................................................................
class Type(FortranBlock):
    #..........................................................
    # pyparsing grammar
    #..........................................................
    BLOCKTYPE = 'Type'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    ABSTRACT = pp.CaselessKeyword('abstract')
    EXTENDS = pp.Suppress(pp.CaselessKeyword('extends')) +\
        pp.Suppress('(') + (VALIDNAME | TEMPLATE_DECL)('name') + pp.Suppress(')')
    BIND = pp.Suppress(pp.CaselessKeyword('bind')) +\
        pp.Suppress('(') + pp.Word(pp.alphas)('name') + pp.Suppress(')')
    TYPE_ATTR_SPEC = \
        pp.delimitedList(EXTENDS('extends') | ABSTRACT | BIND('bind') | PUBLIC |
                         PRIVATE)
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        pp.Optional(',' + TYPE_ATTR_SPEC )('type_attr_spec') +\
        pp.Suppress(pp.Optional('::')) +\
        (pp.NotAny(pp.CaselessKeyword('is'))) +\
        VALIDNAME('name') +\
        pp.Suppress(pp.Optional(ARGLIST)) +\
        pp.Optional(';') + pp.StringEnd()
    #................................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return Type([tokens.name])

    def __init__(self, parameters):
        '''A F90/95/2003 type block:

        type foo
        .
        .
        .
        end type foo

        Input
        -----
        parameters : the groups from the factory static method

        '''
        super(Type, self).__init__(self.BLOCKTYPE, parameters)

        #......................................................................
        # Have to create the SENTINEL grammar because I'm using a different
        # KEYWORD than the name of the class since `type` is a Python keyword
        #......................................................................
        # self.SENTINEL = \
        #     pp.StringStart() +\
        #     pp.CaselessKeyword('end') +\
        #     pp.CaselessKeyword('type') +\
        #     pp.CaselessKeyword(self._name)

#................................................................................
class AutoInterface(FortranBlock):
    #...................................................................
    # pyparsing grammar
    #...................................................................
    BLOCKTYPE = 'AutoInterface'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR =  KEYWORD
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #...................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return AutoInterface(['autointerface'])

    def __init__(self, parameters):
        '''An autointerface block:

        autointerface
        end autointerface

        This block automatically generates interfaces for
        functions/subroutines with 'as' clauses.

        Input
        -----
        parameters : the groups from the factory static method

        '''
        # Send blank line for name to create the proper sentinel
        super(AutoInterface, self).__init__(self.BLOCKTYPE, [''])
        self._name = parameters[0]

    def show(self, mangle=False):
        '''Overloaded show() method.  Finds all functions/subroutines that
        have 'as' naming within its parent block and generates a generic
        interface for it using the as name.  For example:

        function foo(x,y,z)  result(x)  as  bar
        function foo2(x)     result(y)  as  bar

        would generate

        interface bar
           module procedure foo
           module procedure foo2
        end interface

        This is also very useful for templated functions so a consistent
        generic interface can be generated.

        '''
        parent = self._parent
        if parent is not None:
            #................................................................
            # Find all the functions/subroutines that have 'as' names.
            # ................................................................
            funcsAndSubs = parent.findAll((Function, Subroutine))
            funcsAndSubs = [x for x in funcsAndSubs
                            if (hasAsName(('',x)) and not x.isTemplated())]
        else:
            funcsAndSubs = []

        n = tab(self._code[0])

        _dots = '!'+'.'*(80-n)+os.linesep
        block = indent(n, _dots)
        block += indent(n, '! `AutoInterface` Generated Code Section'+os.linesep)
        block += indent(n, _dots)
        block += ''.join(self._code[1:-1])

        # Filter the asNames for funcs/subs that have the same asName
        _asNames = {}
        # for fnc in funcsAndSubs.values():
        for fnc in funcsAndSubs:
            if fnc.asName in _asNames:
                _asNames[fnc.asName].append(fnc.name())
            else:
                _asNames[fnc.asName] = [fnc.name()]

        for k,v in _asNames.items():
            line = indent(n, 'interface %s%s'%(k,os.linesep))
            for name in v:
                line += indent(n+2, 'module procedure %s%s'%(name,os.linesep))

            line += indent(n, 'end interface'+2*os.linesep)
            block += line

        return block+indent(n, _dots)

#................................................................................
class AutoPrivate(FortranBlock):
    #.................................................................
    # pyparsing grammar
    #.................................................................
    BLOCKTYPE = 'AutoPrivate'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR =  KEYWORD
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #.................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return AutoPrivate(['autoprivate'])

    def __init__(self, parameters):
        '''An autoprivate block:

        autoprivate
        end autoprivate

        This block automatically generates private statements for
        functions/subroutines with 'as' clauses.

        Input
        -----
        parameters : the groups from the factory static method

        '''
        # Send blank line for name to create the proper sentinel
        super(AutoPrivate, self).__init__(self.BLOCKTYPE, [''])
        self._name = parameters[0]

    def show(self, mangle=False):
        '''Overloaded show that finds all 'as' name functions/subroutines and
        makes their original names private within the module.

        '''
        parent = self._parent
        if parent is not None:
            #................................................................
            # Find all the functions/subroutines that have 'as' names.
            #................................................................
            funcsAndSubs = parent.findAll((Function, Subroutine))
            funcsAndSubs = [x for x in funcsAndSubs
                            if (hasAsName(('',x)) and not x.isTemplated())]
        else:
            funcsAndSubs = []

        n = tab(self._code[0])

        _dots = '!'+'.'*(80-n)+os.linesep
        block = indent(n, _dots)
        block += indent(n, '! `AutoPrivate` Generated Code Section'+os.linesep)
        block += indent(n, _dots)
        block += ''.join(self._code[1:-1])

        # for fnc in funcsAndSubs.values():
        for fnc in funcsAndSubs:
            line = indent(n, 'private :: %s%s'%(fnc.name(),os.linesep))
            block += line

        return block+indent(n, _dots)

#................................................................................
class AutoUse(FortranBlock):
    #..........................................................
    # pyparsing grammar
    #..........................................................
    BLOCKTYPE = 'AutoUse'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = KEYWORD
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #..........................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return AutoUse(['autouse'])

    def __init__(self, parameters):
        '''An autouse block:

        autouse
        end autouse

        This block automatically generates use statements for
        modules that use types/functions/subroutines in other templated
        modules.

        Input
        -----
        parameters : the groups from the factory static method

        '''
        # Send blank line for name to create the proper sentinel
        super(AutoUse, self).__init__(self.BLOCKTYPE, [''])
        self._name = parameters[0]

    def show(self, mangle=False):
        '''Overloaded show method.  Look for dependencies and put proper name
        mangled use statements.

        '''
        jc = globalSymbolTable.joinCharacter()
        n = tab(self._code[0])

        _dots = '!'+'.'*(80-n)+os.linesep
        block = indent(n, _dots)
        block += indent(n, '! `AutoUse` Generated Code Section'+os.linesep)
        block += indent(n, _dots)
        for line in self._code[ 1:-1 ]:
            self._modLine = line
            ans = self.SHOW_GRAMMARS.searchString( line )
            block += self._modLine
        # block += ''.join( self._code[1:-1] )

        use_decl = map( self.resolveSymbol, self.useDeclarations() )
        use_decl.extend( map( self.resolveSymbol,
                              self._parent.useDeclarations() ) )

        modSet = ASet()
        depSet = ASet( map( self.resolveSymbol,
                            self._parent.dependencies().copy() ) )
        myMod = self.upTo( Module )
        for dep in depSet:
            if myMod is not None:
                if myMod.contains( dep ): continue
            blk = globalSymbolTable.filter( startsWith( dep + jc ) ).values()
            if len( blk ) != 1:
                if len( blk ) == 0 and dep in self._depIgnore: continue
                msg = "WARNING:  Ambiguous or unresolved name in AutoUse block"
                msg += os.linesep
                msg += "Parent block: %s%s"%( self.parent().name(),
                                                                os.linesep)
                msg += "Looking for: %s%s"%(dep, os.linesep)
                msg += "Found: %s%s"%(','.join([b.name() for b in blk]),
                                      os.linesep)
                print msg
                continue
                # raise Exception(msg)

            #................................................................
            # If the block is an instance of a module, then skip it since it
            # was grabbed as a dependency from a use declaration in a child
            # block, i.e. a child block had a "use foo" or something in it
            # which resulted in a module dependency for that block.  These
            # are ignored because we are trying to create "use" declarations
            # for a module based on type/function/subroutine dependencies.
            #................................................................
            blk = blk[0]
            if isinstance( blk, Module ): continue
            mod = blk.upTo( Module )
            if mod is None:
                msg = "Cannot find (%s) in a module for AutoUse" + os.linesep
                msg += "In module (%s) for AutoUse (%s)"
                msg = msg % ( blk.name(), self._parent.name(), str(self) )
                print msg
                continue
                # raise Exception()

            if mod is myMod: continue
            modSet.add( mod.name() )

        modSet.difference_update( use_decl )
        while ( len( modSet ) > 0 ):
            mod = modSet.pop()
            block += indent(n, 'use %s%s'%( mod, os.linesep ) )
        return block+indent( n, _dots )

#................................................................................
class Enumeration(FortranBlock):
    #...................................................................
    # pyparsing grammar
    #...................................................................
    BLOCKTYPE = 'Enumeration'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = pp.StringStart() + KEYWORD + pp.Optional(ARGLIST('args'))
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #...................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        startToken = 0
        increment = 1
        multiple = 1
        try:
            startToken = tokens.args[0]
            increment = tokens.args[1]
            multiple = tokens.args[2]
        except:
            pass
        return Enumeration([startToken,increment,multiple])

    def __init__(self, parameters):
        '''An enumeration block:

        enumeration(start,incr,mult)  |  enumeration start
          INT1 INT2 INT3
        end enumeration

        Results in:

        integer, parameter :: INT1=1*incr+start
        integer, parameter :: INT2=2*incr+start
        integer, parameter :: INT3=3*incr+start

        This block automatically generates integer parameter statements for
        functionals.

        Input
        -----
        parameters : the groups from the factory static method

        '''
        # Send blank line for name to create the proper sentinel
        super(Enumeration, self).__init__(self.BLOCKTYPE,
                                          ['enumeration_'+str(id(self))]+parameters)
        self._starttoken = parameters[0]
        self._increment = parameters[1]
        self._multiple = parameters[2]

    def show(self, mangle=False):
        '''Overloaded show method.  Look for templated dependencies and put
        proper name mangled use statements.

        '''

        block = ''
        try: start = int(self._starttoken)
        except: start = self._starttoken
        try: incr = int(self._increment)
        except: incr = self._increment
        try: mult = int(self._multiple)
        except: mult = self._multiple
        n = tab(self._code[0]) # Tab level of the beginning of the block
        if len(self._code) > 0:
            _n = max(0, tab(self._code[1])-n) # Tab difference for the internal code

        val = incr
        for line in self._code[1:-1]:
            ind = line.find('!')
            if ind < 0: ind = len(line)
            comment = line[ind:]
            for var in " ".join(line[0:ind].split(',')).split():
                block += os.linesep
                block += indent(n, 'integer, parameter :: %s=(%s)+(%d)'%(var,start,val))
                if mult == 1:
                    val += incr
                else:
                    val *= mult
            block += comment + os.linesep
        return block

#................................................................................
class Foreach(FortranBlock):
    #.............................................................
    # pyparsing grammar
    #.............................................................
    BLOCKTYPE = 'Foreach'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = KEYWORD +\
        ARGLIST('args') + pp.Suppress(pp.CaselessKeyword('in')) + ARGLIST2
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #.............................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        replacestrings = tokens.args
        arglist = tokens.args2
        return Foreach(['foreach',replacestrings,arglist])

    def __init__(self, parameters):
        '''A foreach construct:

        foreach(f) in(INT1, INT2, INT3)
          $f$ = $f$ + 1
        end foreach

        Results in:

        INT1 = INT1 + 1
        INT2 = INT2 + 1
        INT3 = INT3 + 1

        Input
        -----
        parameters : the groups from the factory static method

        '''
        # Send blank line for name to create the proper sentinel
        super(Foreach, self).__init__(self.BLOCKTYPE,
                                      [parameters[0]+'_'+str(id(self))]+parameters[1:])
        self._replacestring = parameters[1]
        self._argumentlists = parameters[2]

    def show(self, mangle=False):
        '''Overloaded show method.  Look for templated dependencies and put
        proper name mangled use statements.

        '''

        block = ''

        replist = list()
        for i in self._replacestrings: replist.append(i)
        ln = len(replist)
        nreps = len(self._argumentlists)/ln

        arglist = list()
        for j in xrange(0,nreps):
            tmpargs = list()
            for i in xrange(0,ln):
                tmpargs.append(self._argumentlists[i+ln*j])
            arglist.append(tmpargs)

        print replist,arglist

        n = tab(self._code[0]) # Tab level of the beginning of the block
        if len(self._code) > 0:
            _n = max(0, tab(self._code[1])-n) # Tab difference for the internal code

        repl = self._replacestrings
        for arg in arglist:
            for line in self._code[1:-1]:
                tmpline = line
                for i in xrange(0,len(self._replacestrings)):
                    rs = self._replacestrings[i]
                    ostr = arg[i]
                    tmpline = tmpline.replace('$'+rs+'$', ostr)
                block += indent(n, tmpline + os.linesep)
            block += os.linesep

        return block

#................................................................................
class TestRunner(FortranBlock):
    #.................................................................
    # pyparsing grammar
    #.................................................................
    BLOCKTYPE = 'TestRunner'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        ARGLIST('args')
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #...........................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        args = tokens.args.asList()
        prefix = args[0]
        testArgList = ((len(args) > 1) and args[1:]) or []
        return TestRunner(['TestRunner', prefix, testArgList])

    def __init__(self, parameters):
        '''A test runner block:

        TestRunner(prefix)
        end TestRunner

        This block gathers all functions that start with `prefix` in the
        module found in the sole use statement of the enclosing block.  If no
        use statement is found in the enclosing block, it traverses up its
        block tree to find the enclosing module it is contained in and uses
        that for the function finding.  It then creates the following pattern
        in its show method for each function it finds:

        call setUp(<function name as string here>)
        if(.not. <function with call signature as found in module>) then
           print *,'Test <function name> : failed'
        else
           print *,'Test <function name> : passed'
        endif
        call tearDown(<function name as string here>)

        Input
        -----
        parameters : the groups from the factory static method

        '''
        self._name, self._prefix, self._argList = parameters
        super(TestRunner, self).__init__(self.BLOCKTYPE, parameters)

    def show(self, mangle=False):
        '''Create block of test running code.

        '''
        def _cmp(blkA, blkB):
            '''Comparison function for block names.  Maybe move into Block
            class ???

            '''
            if blkA.name() < blkB.name(): return -1
            if blkA.name() == blkB.name(): return 0
            if blkA.name() > blkB.name(): return 1

        jc = globalSymbolTable.joinCharacter()
        if self._parent:
            _use = self._parent.useDeclarations()
            #.................................................................
            # Three cases:
            # len(_use) > 1  : Enclosing block has more than one use
            #                  statement.  Only expect one in enclosing block.
            # len(_use) == 0 : No use statement in enclosing block.  Search
            #                  for module in family tree.
            # default        : One use statement in enclosing block.  Find it
            #                  in the symbol table and get the functions.
            #.................................................................
            if len(_use) > 1:
                msg = "Enclosing block %s "%(self._parent.name(),)
                msg += "of TestRunner block has more than "
                msg += "one use declaration%s"%(os.linesep,)
                raise Exception(msg)
            elif len(_use) == 0 :
                _mod = self.upTo(Module)
            else:
                _mod = globalSymbolTable.filter(And(startsWith(_use.pop() + jc),
                                              isModule))
                if len(_mod) > 1:
                    msg = "Ambiguous module names found in TestRunner"+os.linesep
                    msg += "Found: %s"%(','.join([m for m in _mod]))
                    raise Exception(msg)

                _mod = _mod.values()[0]

        #....................................................
        # Get the prefixed functions from the _use module
        #....................................................
        _funcs = _mod.children().filter(startsWith(self._prefix)).values()
        _funcs.sort(cmp=_cmp)

        #.....................................................
        # Build code block
        #.....................................................
        n = tab(self._code[0])
        code = ''
        argList = ','.join(self._argList)
        code += indent(n, ''+os.linesep)
        code += indent(n, 'call setUp("__MOD__")'+os.linesep)
        for fnc in _funcs:
            fname = fnc.name()
            code += indent(n, 'call setNTests(this, getNTests(this)+1)'+os.linesep)
            code += indent(n, 'ntest = ntest + 1'+os.linesep)
            code += indent(n, 'err = init(this%shared%messages)'+os.linesep)
            code += indent(n, 'call setUp("%s")%s'%(fname,os.linesep))
            code += indent(n, 'err = init(this%shared%failList_)'+os.linesep)
            code += indent(n, 'ret = %s(%s)%s'%(fname,argList,os.linesep))
            code += indent(n, 'do while(iterate(this%shared%failList_,retval))'+os.linesep)
            code += indent(n, '  if (retval /= 0) ret = .false.'+os.linesep)
            code += indent(n, 'end do'+os.linesep)
            code += indent(n, 'if (.not. ret) then'+os.linesep)
            code += indent(n+3, 'print *, "Test %s... failed"%s'%(fname,os.linesep))
            code += indent(n+3, 'call setNFails(this, getNFails(this)+1)'+os.linesep)
            code += indent(n+3, 'nfail = nfail + 1'+os.linesep)
            code += indent(n, 'else'+os.linesep)
            code += indent(n+3, 'print *, "Test %s... passed"%s'%(fname,os.linesep))
            code += indent(n, 'endif'+os.linesep)
            code += indent(n, 'err = delete(this%shared%failList_)'+os.linesep)
            code += indent(n, 'call tearDown("%s")%s'%(fname,os.linesep))
            code += indent(n, 'err = init(rep,"%s"'%(fname,)+',ret,this%shared%messages)'+os.linesep)
            code += indent(n, 'err = push(replist,rep)'+os.linesep)
            code += indent(n, 'do while (iterate(this%shared%messages,mess))'+os.linesep)
            code += indent(n, '  err = push(messagestack,mess)'+os.linesep)
            code += indent(n, 'end do'+os.linesep)
            code += os.linesep

        code += indent(n, 'call tearDown("__MOD__")'+os.linesep)

        return code

#................................................................................
class ParallelTestRunner(FortranBlock):
    #.............................................................
    # pyparsing grammar
    #.............................................................
    BLOCKTYPE = 'ParallelTestRunner'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = \
        pp.StringStart() +\
        KEYWORD +\
        ARGLIST('args')
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #.............................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        args = tokens.args.asList()
        prefix = args[0]
        testArgList = ((len(args) > 1) and args[1:]) or []
        return ParallelTestRunner(['ParallelTestRunner', prefix, testArgList])

    def __init__(self, parameters):
        '''A test runner block:

        ParallelTestRunner(prefix)
        end ParallelTestRunner

        This block gathers all functions that start with `prefix` in the
        module found in the sole use statement of the enclosing block.  If no
        use statement is found in the enclosing block, it traverses up its
        block tree to find the enclosing module it is contained in and uses
        that for the function finding.  It then creates the following pattern
        in its show method for each function it finds:

        call setUp(<function name as string here>)
        if(.not. <function with call signature as found in module>) then
           print *,'Test <function name> : failed'
        else
           print *,'Test <function name> : passed'
        endif
        call tearDown(<function name as string here>)

        Input
        -----
        parameters : the groups from the factory static method

        '''
        self._name, self._prefix, self._argList = parameters
        super(ParallelTestRunner, self).__init__(self.BLOCKTYPE, parameters)

    def show(self, mangle=False):
        '''Create block of test running code.

        '''
        def _cmp(blkA, blkB):
            '''Comparison function for block names.  Maybe move into Block
            class ???

            '''
            if blkA.name() < blkB.name(): return -1
            if blkA.name() == blkB.name(): return 0
            if blkA.name() > blkB.name(): return 1

        jc = globalSymbolTable.joinCharacter()
        if self._parent:
            _use = self._parent.useDeclarations()
            #.................................................................
            # Three cases:
            # len(_use) > 1  : Enclosing block has more than one use
            #                  statement.  Only expect one in enclosing block.
            # len(_use) == 0 : No use statement in enclosing block.  Search
            #                  for module in family tree.
            # default        : One use statement in enclosing block.  Find it
            #                  in the symbol table and get the functions.
            #.................................................................
            if len(_use) > 1:
                msg = "Enclosing block %s "%(self._parent.name(),)
                msg += "of ParallelTestRunner block has more than "
                msg += "one use declaration%s"%(os.linesep,)
                raise Exception(msg)
            elif len(_use) == 0 :
                _mod = self.upTo(Module)
            else:
                _mod = globalSymbolTable.filter(And(startsWith(_use.pop() + jc),
                                              isModule))
                if len(_mod) > 1:
                    msg = "Ambiguous module names found in ParallelTestRunner"+os.linesep
                    msg += "Found: %s"%(','.join([m for m in _mod]))
                    raise Exception(msg)

                _mod = _mod.values()[0]

        #........................................................
        # Get the prefixed functions from the _use module
        #........................................................
        _funcs = _mod.children().filter(startsWith(self._prefix)).values()
        _funcs.sort(cmp=_cmp)

        #...................................................................
        # Build code block
        #...................................................................
        n = tab(self._code[0])
        code = ''
        argList = ','.join(self._argList)
        code += indent(n, ''+os.linesep)
        code += indent(n, 'call setUp("__MOD__")'+os.linesep+os.linesep)
        for fnc in _funcs:
            fname = fnc.name()
            code += indent(n, 'call setNTests(this, getNTests(this)+1)'+os.linesep)
            code += indent(n, 'ntest = ntest + 1'+os.linesep)
            code += indent(n, 'err = init(this%shared%messages)'+os.linesep)
            code += indent(n, 'err = init(this%shared%failList_)'+os.linesep)
            code += indent(n, os.linesep)
            code += indent(n, 'call setUp("%s")%s'%(fname,os.linesep))
            code += indent(n, 'ret = %s(%s)%s'%(fname,argList,os.linesep))
            code += indent(n, 'call tearDown("%s")%s'%(fname,os.linesep))
            code += indent(n, 'do while(iterate(this%shared%failList_,retval))'+os.linesep)
            code += indent(n+3, 'if (retval /= 0) ret = .false.'+os.linesep)
            code += indent(n, 'end do'+os.linesep)
            code += indent(n, 'localfails = merge(0, 1, ret)'+os.linesep)
            code += indent(n, 'globalfails = 0'+os.linesep)
            code += indent(n, 'call mpi_barrier( testcomm, ierror )'+os.linesep)
            code += indent(n, 'call mpi_reduce( localfails, globalfails, 1, MPI_INTEGER, MPI_SUM, 0, testcomm, ierror )'+os.linesep)
            #
            # Reporting section
            #
            code += indent(n, 'if ( rnk == 0 ) then'+os.linesep)
            # If this proc is the master of the test communicator, it should handle printing
            # messages and also should create the suite report.
            # There should also be a loop over the other processors to collect messages...
            code += indent(n+3, 'if (globalfails > 0) then'+os.linesep)
            code += indent(n+6, 'print *, "Test %s... failed"%s'%(fname,os.linesep))
            code += indent(n+6, 'call setNFails(this, getNFails(this)+1)'+os.linesep)
            code += indent(n+6, 'nfail = nfail + 1'+os.linesep)
            code += indent(n+3, 'else'+os.linesep)
            code += indent(n+6, 'print *, "Test %s... passed"%s'%(fname,os.linesep))
            code += indent(n+3, 'end if'+os.linesep)
            code += indent(n+3, 'err = init(rep,"%s"'%(fname,)+',ret,this%shared%messages)'+os.linesep)
            code += indent(n+3, 'err = push(replist,rep)'+os.linesep)
            code += indent(n+3, 'do while (iterate(this%shared%messages,mess))'+os.linesep)
            code += indent(n+6, 'err = push(messagestack,mess)'+os.linesep)
            code += indent(n+3, 'end do'+os.linesep)
            code += indent(n+3, 'do i = 1, sz-1, 1'+os.linesep)
            code += indent(n+6, '!call mpi_recv(msize, 1, MPI_INTEGER, i, 42, testcomm, stat, ierror)'+os.linesep)
            code += indent(n+6, '!if (msize > 0) then'+os.linesep)
            code += indent(n+6, 'call mpi_probe(i, 43, testcomm, stat, ierror)'+os.linesep)
            code += indent(n+6, 'call mpi_get_count(stat, MPI_CHARACTER, msize, ierror)'+os.linesep)
            code += indent(n+6, 'allocate(cbuff(msize))'+os.linesep)
            code += indent(n+6, 'call mpi_recv(cbuff, msize, MPI_CHARACTER, i, 43, testcomm, stat, ierror)'+os.linesep)
            code += indent(n+6, 'err = init(comb,cbuff)'+os.linesep)
            code += indent(n+6, 'deallocate(cbuff)'+os.linesep)
            code += indent(n+6, 'if (msize > 0) then'+os.linesep)
            code += indent(n+6, 'tmplist = split(comb, sepstr)'+os.linesep)
            code += indent(n+6, 'do while (iterate(tmplist,mess))'+os.linesep)
            code += indent(n+9, 'newmess = ": " + mess'+os.linesep)
            code += indent(n+9, 'mess = stoi(i)'+os.linesep)
            code += indent(n+9, 'mess = mess + newmess'+os.linesep)
            code += indent(n+9, 'newmess = "processor " + mess'+os.linesep)
            code += indent(n+9, 'err = push(messagestack,newmess)'+os.linesep)
            code += indent(n+9, 'err = push(this%shared%messages,newmess)'+os.linesep)
            code += indent(n+6, 'end do'+os.linesep)
            code += indent(n+6, 'end if !(msize > 0)'+os.linesep)
            code += indent(n+3, 'end do'+os.linesep)
            code += indent(n, 'else'+os.linesep)
            # If this proc is not the master of the test communicator, it should send
            # all messages to the master.
            code += indent(n+3, 'comb = join(sep, this%shared%messages)'+os.linesep)
            code += indent(n+3, 'msize = length(comb)'+os.linesep)
            code += indent(n+3, '!call mpi_send(msize, 1, MPI_INTEGER, 0, 42, testcomm, ierror)'+os.linesep)
            code += indent(n+3, 'if (msize > 0) then'+os.linesep)
            code += indent(n+6, 'cbuff => getText(comb)'+os.linesep)
            code += indent(n+6, 'call mpi_send(cbuff, msize, MPI_CHARACTER, 0, 43, testcomm, ierror)'+os.linesep)
            code += indent(n+6, 'nullify(cbuff)'+os.linesep)
            code += indent(n+3, 'else'+os.linesep)
            code += indent(n+6, 'call mpi_send(\' \', 0, MPI_CHARACTER, 0, 43, testcomm, ierror)'+os.linesep)
            code += indent(n+3, 'end if !(msize > 0)'+os.linesep)
            code += indent(n, 'end if'+os.linesep)
            code += indent(n, 'err = delete(this%shared%failList_)'+os.linesep)
            code += os.linesep

        code += indent(n, 'call tearDown("__MOD__")'+os.linesep)

        return code

#................................................................................
class AutoDelete(FortranBlock):
    #..............................................................
    # pyparsing grammar
    #..............................................................
    BLOCKTYPE = 'AutoDelete'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = \
        KEYWORD +\
        ARGLIST('args')
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #..............................................................
    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        name = tokens.args[0]
        return AutoDelete(['autodelete_'+name])

    def __init__(self, parameters):
        '''An autodelete block:

        AutoDelete(name)
          type (DLList<integer>) :: a,b,c
        end AutoDelete

        This block will store all the symbols of the type() declarations,
        i.e. a,b,c above for later use with the GarbageCollect block.

        Input
        -----
        parameters : the groups from the factory static method

        '''
        super(AutoDelete, self).__init__(self.BLOCKTYPE, parameters)
#         self._symbols = []

    def show(self, mangle=False):
        '''Find all the derived type names in the block and store them for the
        GarbageCollect block.

        '''

        block = ''
        n = tab(self._code[0])
        untab = tab(self._code[1])-n
        parent = self.parent()

        for line in self._code[1:-1]:
            ans = TEMPLATE_TYPE.searchString(line)
            try:
                name = ans[0].name
                newName = self.resolveSymbol(name)
                block += line.replace(name, newName)[untab:]
            except:
                raise Exception('''Could not figure out line: %s %sIn AutoDelete%s'''%(line,os.linesep,os.linesep))

#         for line in self._code[1:-1]:
#             ans = TEMPLATE_TYPE.searchString(line)
#             try:
#                 name = ans[0].name
#                 symbols.extend(ans[0].symbolList.asList())
#                 newName = self.resolveSymbol(name)
#                 block += line.replace(name, newName)[untab:]
#             except:
#                 raise Exception('''Could not figure out line: %s %sIn
#             AutoUse%s'''%(line,os.linesep,os.linesep))

        return block

#     def getSymbols(self):
#         return self._symbols

#................................................................................
class GarbageCollect(FortranBlock):
    #............................................................
    # pyparsing grammar
    #............................................................
    BLOCKTYPE = 'GarbageCollect'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = \
        KEYWORD +\
        ARGLIST('args')
    SENTINEL = \
        pp.StringStart() +\
        pp.CaselessKeyword('end') +\
        KEYWORD
    #.............................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        name = tokens.args[0]
        errorToken = tokens.args[1]
        return GarbageCollect(['gc_'+name, name, errorToken])

    def __init__(self, parameters):
        '''An GarbageCollect that works in concert with the AutoDelete block.
        It autogenerates calls to the delete(...) function for symbols found
        in the AutoDelete block.  For instance, if you have the following auto
        delete block:

        AutoDelete(locals)
          type (DLList<integer>) :: a,b,c
        end AutoDelete

        Then, a GarbageCollect block like this:

        GarbageCollect(locals, lerr)
        end GarbageCollect

        will generate:

        lerr = catch(delete(a), .true.)
        lerr = catch(delete(b), .true.)
        lerr = catch(delete(c), .true.)


        Input
        -----
        parameters : the groups from the factory static method

        '''
        super(GarbageCollect, self).__init__(self.BLOCKTYPE, parameters)
        self._autoDelName = parameters[1]
        self._err = parameters[2]

    def show(self, mangle=False):
        '''Work with the AutoDelete block to auto create delete calls.

        '''

        block = ''
        name = self._autoDelName
        err = self._err
        n = tab(self._code[0])
        if len(self._code) > 0:
            _n = max(0, tab(self._code[1])-n)

        autoDelBlocks = self._parent.findAll((AutoDelete,), depth=1)
        auto = [x for x in autoDelBlocks
                if x.name().startswith('autodelete_'+name)]

        # A Specialization block may contain an AutoDelete
        specBlocks = self._parent.findAll((Specialization,), depth=1)
        for x in specBlocks:
            auto.extend([y for y in x.findAll((AutoDelete,), depth=1)
                         if y.name().startswith('autodelete_'+name)])

        # auto = self.parent().children().filter(startsWith('autodelete_'+name)).values()
        if len(auto) > 1:
            print 'Found multiple autodelete blocks:'
            for x in auto:
                print '-> %s'%(x.name(),)
            raise Exception('More than one autodelete block found in parent!')
        elif len(auto) == 0:
            return block

        auto = auto[0]
#         for sym in auto.getSymbols():
#             block += indent(n, err+' = catch(delete(%s), .true.)%s'%(sym,os.linesep))

        for sym in auto.symbols:
            block += indent(n, err+' = catch(delete(%s), .true.)%s'%(sym,os.linesep))

        for line in self._code[1:-1]:
            block += line[_n:]

        return block

#................................................................................
class Specialization(FortranBlock):
    BLOCKTYPE = 'Specialization'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    GRAMMAR = KEYWORD+pp.LineEnd()
    SENTINEL = \
        pp.CaselessKeyword('end') +\
        KEYWORD
    SPECIAL = \
        pp.Suppress(pp.CaselessKeyword('template')) +\
        pp.Suppress(LEFT_TEMPTOK) + \
        pp.Optional(pp.Combine(TEMPLATE_PARAM_LIST, adjacent=False))('spec') +\
        pp.Suppress(RIGHT_TEMPTOK)

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return Specialization(['special'])

    def __init__(self, parameters):
        '''This is a specialization block which should only appear inside a
        template block.  It is used to produce code that appears within it for
        certain template parameters.  This is an example:

        template <class Object, class Element>

        Specialization
        template <real, real>
        function foo(bar, a,b,c) result(blah)
           type (myType<real, real>) :: bar
           real :: a,b
           integer :: c

           bar%a = bar%a * a
           bar%b = bar%b * b
        end function foo

        template <>
        function foo(bar, a,b,c) result(blah)
           type (myType<Object, Element>) :: bar
           real :: a,b
           integer :: c
           print *, 'This is the default'
        end function foo
        end Specialization

        end template

        The above Specialization block will only show the first function if Object is
        real and Element is real, but shows the second function default
        (i.e. Using <>) otherwise.

        Input
        -----
        parameters : the groups from the factory static method

        '''

        super(Specialization, self).__init__(self.BLOCKTYPE, parameters)
        self._name = parameters[0]+'_'+str(id(self))
        self._specLists = []
        self._codeBuff = ''
        self._default = None
        self._SPECIAL = self.SPECIAL + pp.SkipTo(self.SPECIAL|self.SENTINEL)
        self._SPECIAL.setParseAction(self._addSpec)

    def _addSpec(self, s, l, tokens):
        '''Create specialization list where each entry is a tuple of template
        parameters tuple and corresponding code.

        '''
        end = pp.getTokensEndLoc()
        code = s[l:end].splitlines(True)
        code.append(os.linesep) # append carriage return
        if tokens.spec == '':
            self._default = code[1:]
            spec = ('',)
        else:
            spec = templateSpecToTuple(tokens.spec)
        self._specLists.append((spec, code[1:]))

    def setState(self, state):
        '''Overloaded to select which code block to use.  After the fist call,
        a code block is selected.  Subsequent calls can be used to switch to
        others.

        '''
        self._state = state
        if state is None:
            self._state = None
#             if len(self._specLists) == 0:
#                 self._code = self._codeBuff.splitLines()
            return

        self._code = []
        for spec,code  in self._specLists:
            if all([(x[1] == y) for x,y in zip(self._state, spec) if y != '*']):
                self._code = code
                return

        if self._default:
            self._code = self._default

    def addLine(self, line, n=None):
        '''Overloaded addLine to add to code buffer and parse later in
        sentinel method.

        '''
        self._codeBuff += line

    def show(self, mangle=False):
        '''Overloaded show to break _codeBuff into code lines.

        '''
        self._code = self._codeBuff.splitlines(True)
        return super(Specialization, self).show(mangle)

    def sentinel(self, line):
        '''Overloaded sentinel method to grab all the specialization blocks.

        '''
        if self.SENTINEL.searchString(line):
            self.addLine(line)
            self._SPECIAL.searchString(self._codeBuff)
            return True

    def __getstate__(self):
        '''Builds a dictionary tree of the block.

        '''
        _dd = self.__dict__.copy()
        _dd.pop('_SPECIAL', None)
        return super( Specialization, self ).__getstate__( _dd )

    def serialize( self ):
        return self.__getstate__()

#................................................................................
class Instantiate( FortranBlock ):
    BLOCKTYPE = 'Instantiate'
    KEYWORD = pp.Suppress(pp.CaselessKeyword(BLOCKTYPE))
    OPTNAME = pp.Optional(VALIDNAME)('name')
    GRAMMAR = KEYWORD + OPTNAME + pp.LineEnd()
    SENTINEL = \
        pp.CaselessKeyword('end') +\
        KEYWORD

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        name = tokens.name
        return Instantiate( [ 'instantiate_' + tokens.name ] )

    def __init__(self, parameters):
        '''An instantiate block is simply used to instantiate template symbols
        listed in the block.  This can be used to separate the build cycle
        into phases.  Phase I would be to have instantiate blocks build all
        the templates for you and put them somewhere.  Phase II would be to
        read the sources and mangle the template names.  Phase III would be to
        parse the headers of all the files and build the dependencies for a
        Makefile. Also, these blocks can be placed in a file directly to
        trigger instantiations.  This can be useful if you place functions
        within a templated module and use an `as` name to wrap them in generic
        interfaces.  You can the simply use the `as` name in your code instead
        of a template call.
        '''

        super(Instantiate, self).__init__(self.BLOCKTYPE, parameters)
        self._name = parameters[0]+'_'+str(id(self))

    def show( self, mangle ):
        return ''

#................................................................................
class Pod(FortranBlock):
    BLOCKTYPE = 'Pod'
    podWord = BLOCKTYPE
    #.....................................................................
    # pyparsing grammar
    #.....................................................................
    KEYWORD = pp.CaselessKeyword(BLOCKTYPE)
    VALIDNAME = pp.Word(pp.alphas, pp.alphanums+'_')
    FROM = pp.CaselessKeyword('from')

    GRAMMAR = \
        pp.Suppress(KEYWORD) +\
        VALIDNAME('podName') +\
        pp.Suppress(FROM) +\
        VALIDNAME('modName')
    #.....................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return Pod(tokens.asList())

    def __init__(self, parameters):
        '''A (P)ython (O)n (D)emand block:

        pod foo from barmodule
        .
        .
        .
        end pod foo

        The code contained within the block is processed AT SHOW TIME.  Use
        the Prepod block for processing at parse time.

        Input
        -----
        parameters : the groups from the class regex
        table : the global symbol table to store the block in

        '''
        super(Pod, self).__init__(self.BLOCKTYPE, parameters)
        self._name = parameters[0].partition('~')[0]+'~'+str(id(self))

        # Local parser and scanner
        self._scan = LineScanner()
        self._module = None
        self._fnc = None
        self._parser = FortranBlockParser(self._scan)
        self._parser.register([Template,
                               Module,
                               Function,
                               Subroutine,
                               Type,
                               Interface,
                               Pod,
                               Program,
                               AutoInterface,
                               AutoPrivate,
                               AutoUse])

    def name(self):
        # No mangled names for interface names within templates
        return self._name

    def show(self, mangle):
        '''Overloaded show() for Pods.

        '''
        self._module = __import__(self._parameters[1])
        self._fnc = getattr(self._module, self._parameters[0].partition('~')[0], None)
        if self._fnc is None:
            raise Exception("Could not find pod: %s"%(str(self._parameters[0]),))

        # Strip all leading spaces from code block and indent to proper Python
        # Strip the 'pod foo ...' and 'end pod' lines
        code = ''
        for line in self._code[1:]:
            if line.rstrip() != '':
                nspace = tab(line) # indentation of Python code
                break
        code = ''.join(map(lambda s: s[nspace:], self._code[1:-1]))

        #......................................................................
        # Use the local line scanner, symbol table as a copy of
        # globalSymbolTable and block parser on the returned code string.
        # Parse it and call show on the returned block from the parser.  This
        # will perform all the necessary substitutions and name mangling.
        #......................................................................
        nspace = tab(self._code[0])
        code = self._fnc(code, globals())
        code = (os.linesep).join(map(lambda x: indentNoStrip(nspace, x),
                                     code.splitlines()))
        code += os.linesep
        self._scan.loadString(code, self.name())
        blk = self._parser.parse(AFile(parameters=[self._scan.name(),
                                                   self._scan.hash()]))

        _dots = indent(nspace, '!'+'.'*(80-nspace)+os.linesep)
        _banner = _dots
        _banner += indent(nspace,'! Pod %s Generated Code Block%s'%(self._name,os.linesep))
        _banner += _dots
        return _banner + blk.show(mangle) + _dots

    def __getstate__(self):
        _dd = self.__dict__.copy()
        _dd.pop('_fnc', None)
        _dd.pop('_module', None)
        _dd.pop('_parser', None)
        _dd.pop('_scan', None)
        return super( Pod, self ).__getstate__( _dd )

    def serialize(self):
        '''Builds a dictionary tree of the block.

        '''
        return self.__getstate__()

#................................................................................
class Prepod(FortranBlock):
    BLOCKTYPE = 'Prepod'
    podWord = BLOCKTYPE
    #.....................................................................
    # pyparsing grammar
    #.....................................................................
    KEYWORD = pp.CaselessKeyword(BLOCKTYPE)
    VALIDNAME = pp.Word(pp.alphas, pp.alphanums+'_')
    FROM = pp.CaselessKeyword('from')

    GRAMMAR = \
        pp.Suppress(KEYWORD) +\
        VALIDNAME('podName') +\
        pp.Suppress(FROM) +\
        VALIDNAME('modName')
    #.....................................................................

    @staticmethod
    def factory(s, lineNo, tokens):
        '''Factory method called by the BlockParser to produce an
        instance of the class.

        '''
        return Prepod(tokens.asList())

    def __init__(self, parameters):
        '''A (P)ython (O)n (D)emand block:

        prepod foo from barmodule
        .
        .
        .
        end prepod foo

        The code contained within the block is processed AT PARSING TIME.

        Input
        -----
        parameters : the groups from the class regex
        table : the global symbol table to store the block in

        '''
        super(Prepod, self).__init__(self.BLOCKTYPE, parameters)
        self._name = parameters[0]+'_'+str(id(self))
        self._module = __import__(self._parameters[1])
        self._fnc = getattr(self._module, parameters[0], None)

        # Local parser and scanner
        self._scan = LineScanner()
        self._parser = FortranBlockParser(self._scan)
        self._parser.register([Template,
                               Module,
                               Function,
                               Subroutine,
                               Type,
                               Interface,
                               Program,
                               AutoInterface,
                               AutoPrivate,
                               AutoUse])

    def name(self):
        # No mangled names for interface names within templates
        return self._name

    def sentinel(self, line):
        '''Overloaded sentinel() for prepods.

        '''
        if not self.SENTINEL.searchString(line): return False
        if self._fnc is None:
            raise Exception("Could not find prepod: %s"%(str(self._fnc,)))

        # Strip all leading spaces from code block and indent to proper Python
        # Strip the 'prepod ...' and 'end prepod' lines
        code = ''
        _n = tab(self._code[0])           # indentation of Prepod block
        for line in self._code[1:]:
            if line.rstrip() != '!':
                nspace = tab(line[1:]) + 1 # indentation of Python code
                break
        # nspace = tab(self._code[1][1:])+1 # indentation of Python code
        code = ''.join(map(lambda s: s[nspace:], self._code[1:]))

        #......................................................................
        # Logic:
        #
        # Use the local line scanner, a new symbol table and local block
        # parser on the returned code string.  Load the code string into the
        # scanner, clear the prepod code string and call the block parser with
        # self as the starting block.
        #......................................................................
        self._scan.loadString(self._fnc(code, globals()), self.name())
        self._code = []
        self._parser.parse(self)
        self._code = map(lambda s: indent(_n, s), self._code)

        _dots = indent(_n, '!'+'.'*(80-nspace)+os.linesep)
        _banner = _dots
        _banner += indent(_n,'! Prepod %s Generated Code Block%s'%(self._name,os.linesep))
        _banner += _dots
        self._code.insert(0,_banner)
        self._code.append(os.linesep+_dots)

        return True

    def __getstate__(self):
        _dd = self.__dict__.copy()
        _dd.pop('_fnc', None)
        _dd.pop('_module', None)
        _dd.pop('_parser', None)
        _dd.pop('_scan', None)
        return super( Prepod, self ).__getstate__(_dd)

    def serialize(self):
        '''Builds a dictionary tree of the block.

        '''
        return self.__getstate__()

#................................................................................
class PyF95Directives( FortranBlock ):
    BLOCKTYPE = 'PyF95Directives'
    #.....................................................................
    # pyparsing grammar
    #.....................................................................
    GRAMMAR = PYF95_START
    SENTINEL = PYF95_END

    @staticmethod
    def factory( s, l, t):
        blk = PyF95Directives( [ PyF95Directives.BLOCKTYPE ] )
        return blk

    def __init__( self, parameters ):
        super( PyF95Directives, self ).__init__( self.BLOCKTYPE, parameters )
        self._name = parameters[ 0 ]
        self.external = []
        self.alwayswrite = False

        # Override the DEP_GRAMMAR set in FortranBlock
        self.DEP_GRAMMAR = \
            ( PYF95_EXTERNAL.copy() ).setParseAction( self.__external ) | \
            ( PYF95_ALWAYSWRITE.copy() ).setParseAction( self.__alwayswrite )

    def sentinel( self, line ):
        if self.SENTINEL.searchString( line ):
            self.addLine( line )
            return True
        return False

    def __external( self, s, l, t ):
        self.external.append( t.symbol )

    def __alwayswrite( self, s, l, t ):
        setattr( self, 'alwayswrite', ( t.bool.lower() == "true" ) )

    def show( self, mangle=False ):
        return ''.join( self._code ).replace( '#', '!' )

