#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
# Copyright (c) 2009,2010,2011 Michael List, michael.list@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
from blockit.parsers import pyp as pp
# pp.ParserElement.enablePackrat()

#................................................................................
# pyparsing Global ParserElement objects used by all blocks
#................................................................................
TEMPLATE_TOKENS = ('<','>')
RESERVED_KEYWORDS = ['template']
LEFT_TEMPTOK, RIGHT_TEMPTOK = map(pp.Literal, TEMPLATE_TOKENS)
COMMENT = pp.Literal( "!" ) + pp.restOfLine
#COMMENT = \
#    pp.originalTextFor( pp.Literal( '!' ) + \
#                        pp.Optional( 
#        pp.OneOrMore( pp.Word( pp.alphanums + '_' ))) +\
#                            pp.lineEnd )
CONTINUATION = pp.Literal( "&" )
INLINE_CONTINUATION = \
    CONTINUATION + \
    pp.Optional( COMMENT )+ \
    CONTINUATION
# VALIDNAME = pp.Word(pp.alphas+'_', pp.alphanums+'_%').ignore( INLINE_CONTINUATION )
VALIDNAME = pp.Word(pp.alphas+'_', pp.alphanums+'_%')
VALIDNAME_WITH_PAREN = VALIDNAME+pp.Optional('('+VALIDNAME+')')
VAR_ASSIGNMENT = VALIDNAME + pp.Literal('=')
NUM = pp.Word(pp.nums)
INTEGER = pp.Combine(pp.Optional('-')+pp.Word(pp.nums))
DECIMAL_NUM = pp.Combine(INTEGER+'.'+pp.Optional(NUM))
ENGINEERING_NUM = pp.Combine((DECIMAL_NUM|INTEGER) +\
                              pp.oneOf('e d', caseless=True) +\
                              INTEGER)

PRIVATE = pp.CaselessKeyword('private')
PUBLIC = pp.CaselessKeyword('public')
PROTECTED = pp.CaselessKeyword('protected')
ACCESS_SPEC = PRIVATE ^ PUBLIC ^ PROTECTED

OUT = pp.CaselessKeyword('out')
IN = pp.CaselessKeyword('in')
INOUT = pp.CaselessKeyword('inout')
INTENT = \
    pp.CaselessKeyword('intent') +\
    pp.Suppress('(') +\
    (IN|OUT|INOUT)('intent') +\
    pp.Suppress(')')

NUM.setParseAction(lambda s,l,t: int(t[0]))
INTEGER.setParseAction(lambda s,l,t: int(t[0]))
DECIMAL_NUM.setParseAction(lambda s,l,t: float(t[0]))

TEMPLATE_DECL = pp.Forward()
VALID_TEMPLATE_PARAM = (pp.Literal('*')|TEMPLATE_DECL|VALIDNAME_WITH_PAREN|INTEGER)
TEMPLATE_PARAM_LIST =  (VALID_TEMPLATE_PARAM + pp.ZeroOrMore(','+VALID_TEMPLATE_PARAM))
TEMPLATE_DECL << \
    VALIDNAME +\
    pp.Combine((LEFT_TEMPTOK + TEMPLATE_PARAM_LIST + RIGHT_TEMPTOK), 
               adjacent=False)

ARGLIST = \
    pp.Group(
    pp.Suppress('(')+\
        pp.Optional(pp.delimitedList(VALIDNAME |
                                     pp.quotedString |
                                     (ENGINEERING_NUM | DECIMAL_NUM | INTEGER)))+\
        pp.Suppress(')'))

# Arglist2 is for things with 2 arg lists. we should probably have a dictionary of
# these to allow multiple arg names.
ARGLIST2 = \
    pp.Group(
    pp.Suppress('(')+\
        pp.Optional(pp.delimitedList(VALIDNAME |
                                     pp.quotedString |
                                     (ENGINEERING_NUM | DECIMAL_NUM | INTEGER)))+\
        pp.Suppress(')'))('args2')

#........................................
# Attribute Spec
#........................................
TARGET = pp.CaselessKeyword('target')
OPTIONAL = pp.CaselessKeyword('optional')
SAVE = pp.CaselessKeyword('save')
POINTER = pp.CaselessKeyword('pointer')
ALLOCATABLE = pp.CaselessKeyword('allocatable')
EXTERNAL = pp.CaselessKeyword('external')
INTRINSIC = pp.CaselessKeyword('intrinsic')
ASYNCHRONOUS = pp.CaselessKeyword('asynchronous')
VALUE = pp.CaselessKeyword('value')
DIMENSION = \
    pp.CaselessKeyword('dimension') +\
    pp.Literal('(') +\
    pp.delimitedList(pp.Literal(':')|pp.Word(pp.nums)) +\
    pp.Literal(')')
ATTR_SPEC = ACCESS_SPEC | INTENT | TARGET | SAVE | POINTER | ALLOCATABLE | \
    EXTERNAL | INTRINSIC | VALUE | DIMENSION | ASYNCHRONOUS

ONLY_CLAUSE = \
    pp.Literal(',')+pp.CaselessKeyword('only')+pp.Literal(':')

#............................................................
# Types
#............................................................
INTEGER = \
    pp.originalTextFor( \
        pp.CaselessKeyword( 'integer' ) +\
        pp.Optional( '(' +\
                     pp.originalTextFor(VALIDNAME) +\
                     ')' \
                   ) \
                      )
LOGICAL = \
    pp.originalTextFor( \
        pp.CaselessKeyword('logical') )
REAL = \
    pp.originalTextFor( \
        pp.CaselessKeyword('real') +\
        pp.Optional( '(' +\
                     pp.originalTextFor(VALIDNAME) +\
                     ')' \
                   ) \
                      )
CHARACTER = pp.CaselessKeyword('character')
INTRINSIC_TYPE = \
    pp.StringStart() +\
    ( INTEGER | REAL | LOGICAL )('name') +\
    pp.Optional(pp.Suppress(',') + pp.delimitedList(ATTR_SPEC)) +\
    pp.Optional(pp.Suppress('::')) +\
    pp.delimitedList(VALIDNAME)('symbolList')
USE_DECL = \
    pp.StringStart() +\
    pp.CaselessKeyword('use')('use') +\
    ( pp.originalTextFor( TEMPLATE_DECL )| VALIDNAME )('name') +\
    pp.Optional(ONLY_CLAUSE) +\
    pp.Optional(pp.delimitedList( pp.originalTextFor( TEMPLATE_DECL ) ^ 
                                  VALIDNAME ) )('onlyList')
TEMPLATE_MOD_PROCEDURE_DECL = \
    pp.StringStart() +\
    pp.Suppress( pp.CaselessKeyword('module procedure') ) + \
    ( pp.originalTextFor( TEMPLATE_DECL ) )('name')
IMPORT_DECL = \
    pp.StringStart() +\
    pp.Suppress( pp.CaselessKeyword('import') ) + \
    ( pp.originalTextFor( TEMPLATE_DECL )| VALIDNAME )('name')
TEMPLATE_TYPE = \
    pp.StringStart() +\
    pp.CaselessKeyword('type') +\
    pp.Suppress('(') +\
    pp.originalTextFor(TEMPLATE_DECL ^ VALIDNAME)('name') +\
    pp.Suppress(')') +\
    pp.Optional(pp.Suppress(',') + pp.delimitedList(ATTR_SPEC)) +\
    pp.Suppress('::') +\
    pp.delimitedList(VALIDNAME)('symbolList')
TEMPLATE_SUB = \
    pp.StringStart() +\
    pp.CaselessKeyword('call') +\
    pp.originalTextFor(TEMPLATE_DECL)('name') +\
    pp.Optional(ARGLIST('args'))
TEMPLATE_FUNC = \
    pp.Suppress( pp.oneOf('= ( ,') ) +\
    pp.originalTextFor(TEMPLATE_DECL)('name') +\
    pp.Optional(ARGLIST('args'))
TEMPLATE_SYMBOL = pp.originalTextFor(TEMPLATE_DECL)('name')

# TEMPLATE_FUNC.ignore( INLINE_CONTINUATION )

#........................................
# PyF95 directives grammars
#........................................
PYF95_KEYWORD = pp.LineStart() + pp.Literal('#PyF95')
PYF95_START = \
    PYF95_KEYWORD + \
    pp.Literal( '.start' )
PYF95_END = \
    PYF95_KEYWORD + \
    pp.Literal( '.end' )
PYF95_EXTERNAL = \
    pp.Suppress( 
      PYF95_KEYWORD + \
      pp.Literal( '.external' ) + \
      pp.Literal( ':' ) ) + \
    ( pp.Combine( TEMPLATE_DECL, adjacent=False ) | VALIDNAME )( 'symbol' )
PYF95_ALWAYSWRITE = \
    pp.Suppress(
      PYF95_KEYWORD + \
      pp.Literal( '.alwayswrite' ) + \
      pp.Literal( ':' ) ) + \
    pp.Word( pp.alphas )('bool')
PYF95_BLOCK = pp.Suppress( PYF95_START ) + pp.SkipTo( PYF95_END )
 
########################################################################
##
## Unit-Tests
##
########################################################################

if __name__ == "__main__":

    import unittest

#=======================================================================
    class TestFGrammar(unittest.TestCase):
#=======================================================================
    
        def setUp(self): pass

        def tearDown(self): pass

    #-------------------------------------------------------------------
        def testUseDecl(self):
    #-------------------------------------------------------------------
            """Checks parsing of use statements"""

            bnf = USE_DECL

            # The USE_DECL should only parse 1 occurrance of the match because
            # it uses StringStart method.
            ctr = 0
            for match,s,e in bnf.scanString( """
                    use common_area, only: var1,var2, var3
                    use DLList<String>
                """ ):
                ctr += 1
            self.failUnlessEqual(ctr,1,"Parsed more than one use statement!")

            # Make sure it peels out the use and name correctly for a template
            try:
                output = bnf.parseString("use DLList<String>")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(output['use'], 'use', "Match use not correct")
            self.failUnlessEqual(output['name'], 'DLList<String>', "Match name not correct")

            # Make sure the only statement works
            try:
                output = bnf.parseString(" use common_area, only: var1,var2, var3")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(output['use'], 'use',"Match use not correct")
            self.failUnlessEqual(output['name'], 'common_area', "Match name not correct")
            self.failUnlessEqual(len(output['onlyList']), 3, "Match list len not correct")
            for i in xrange(1,4):
                self.failUnlessEqual(output['onlyList'][i-1], 'var'+str(i), "Match list varname not correct")

        def testIntent(self):
            """Checks intent parsing"""

            bnf = INTENT

            try:
                output = bnf.parseString("intent (inout)")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(output['intent'], 'inout', 'Improper parse of intent')

            try:
                output = bnf.parseString("intent (in)")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(output['intent'], 'in', 'Improper parse of intent')

            try:
                output = bnf.parseString("intent (out)")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(output['intent'], 'out', 'Improper parse of intent')


    suite = unittest.TestLoader().loadTestsFromTestCase(TestFGrammar)
    unittest.TextTestRunner(verbosity=2).run(suite)
