#................................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
# Copyright (c) 2009,2010,2011 Michael List, michael.list@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
__doc__ = """
(C) 2010,2011 Michael List, michael.list@gmail.com
(C) 2010,2011 David Car, david.car7@gmail.com
"""

import external.pyparsing as pp

VALIDNAME = pp.Word(pp.alphas+'_', pp.alphanums+'_')

NUM = pp.Word(pp.nums)
INTEGER = pp.Combine((pp.Optional('-')|pp.Optional('+'))+pp.Word(pp.nums))
DECIMAL_NUM = pp.Combine(INTEGER+'.'+pp.Optional(NUM)+pp.Suppress(pp.Optional('f')|pp.Optional('F')))
HEX_NUM = pp.Combine( pp.Literal( '0x' ) + pp.Word( pp.nums + 'abcdefABCDEF' ))
ENGINEERING_NUM = pp.Suppress(pp.Optional('(')) + pp.Combine( (DECIMAL_NUM | INTEGER) + \
    pp.oneOf( 'e d', caseless=True ) + pp.Optional( pp.oneOf('+ -') ) + NUM ) + \
    pp.Suppress(pp.Optional(')'))
BIN_NUM = pp.Word('01')

#NUM.setParseAction(lambda s,l,t: int(t[0]))
#INTEGER.setParseAction(lambda s,l,t: int(t[0]))
#DECIMAL_NUM.setParseAction(lambda s,l,t: float(t[0]))
#ENGINEERING_NUM.setParseAction(lambda s,l,t: float(t[0]))

def isbin(instr):
    try:
        out = BIN_NUM.parseString(instr)
        return True
    except:
        return False

def isoct(instr):
    return False

def ishex(instr):
    try:
        out = HEX_NUM.parseString(instr) 
        return True
    except:
        return False

UNIXFILE = pp.Group(pp.OneOrMore(pp.Literal("/") + pp.Word(pp.alphanums + '_-.')))

QUOTESTRING = pp.QuotedString(quoteChar='"')

if __name__ == "__main__":

    import unittest

#=======================================================================
    class TestBaseGrammar(unittest.TestCase):
#=======================================================================

        def setUp(self): pass
        def tearDown(self): pass

    #-------------------------------------------------------------------
        def testIsHex(self):
    #-------------------------------------------------------------------
            """Testing the ishex function"""

            self.assertTrue(ishex('0xff'), "Failed to parse hex number")
            self.assertFalse(ishex('33'), "Parsed non-hex number")
            self.assertFalse(ishex('ab'), "Parsed non-hex number")


    #-------------------------------------------------------------------
        def testIsBin(self):
    #-------------------------------------------------------------------
            """Testing the isbin function"""

            self.assertTrue(isbin('0101010'), "Failed to parse bin number")
            self.assertFalse(isbin('33'), "Parsed non-bin number")
            self.assertFalse(isbin('ab'), "Parsed non-bin number")
            self.assertFalse(isbin('0x01'), "Parsed non-bin number")


    #-------------------------------------------------------------------
        def testFortranHexNum(self):
    #-------------------------------------------------------------------
            """Testing parsing of a Fortran hex number"""

            try:
                output = HEX_NUM.parseString("z'0a'")
            except:
                self.fail("Failed to parse")


    #-------------------------------------------------------------------
        def testEngineeringNum(self):
    #-------------------------------------------------------------------
            """Testing parsing of an engineering number"""

            try:
                output = ENGINEERING_NUM.parseString("-1.0e+04")
            except:
                self.fail("Failed to parse")

            try:
                output = ENGINEERING_NUM.parseString("-1.0e-0")
            except:
                self.fail("Failed to parse")


    #-------------------------------------------------------------------
        def testUnixFile(self):
    #-------------------------------------------------------------------
            """Testing parsing of unix style file & path names"""
            try:
                output = UNIXFILE.parseString("/usr/local/cuda/include/driver_types.h")
            except:
                self.fail("Failed to parse")


    suite = unittest.TestLoader().loadTestsFromTestCase(TestBaseGrammar)
    unittest.TextTestRunner(verbosity=2).run(suite)
