#................................................................................
# The MIT License
#
# Copyright (c) 2009-2010 David Car, david.car7@gmail.com
# Copyright (c) 2009-2010 Michael List, michael.list@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#................................................................................

__doc__ = """
(C) 2010 Michael List, michael.list@gmail.com
(C) 2010 David Car, david.car7@gmail.com
"""

from blockit.parsers import pyp as pp
import blockit.basegrammar as gram


#GENERALLINE = pp.Group( pp.OneOrMore( pp.Word( pp.printables ) ) )

CPPCOMMENT = pp.Literal("//") + pp.restOfLine
CCOMMENT = pp.cStyleComment
FCOMMENT = pp.Literal( "!" ) + pp.restOfLine

LINES = pp.Combine(pp.OneOrMore( pp.Word( pp.printables ) )) + pp.restOfLine

POUND = pp.Literal( '#' ).suppress()

# num file num num ...
POUNDLINE = POUND + pp.restOfLine
POUNDNUMFILE = pp.lineStart + POUND + gram.NUM + pp.QuotedString("\"", multiline=False) + pp.Group(pp.restOfLine) + pp.lineEnd
POUNDNUMFILE = pp.lineStart + POUND + gram.NUM + pp.QuotedString("\"", multiline=False) + pp.ZeroOrMore(gram.NUM)

#if
IF = pp.CaselessKeyword('if')
POUNDIF = pp.LineStart() + pp.Group( POUND + IF + pp.Group( pp.restOfLine ) )
POUNDIF.ignore( CPPCOMMENT )
POUNDIF.ignore( CCOMMENT )
POUNDIF.ignore( FCOMMENT )

#else
ELSE = pp.CaselessKeyword('else')
POUNDELSE = pp.LineStart() + POUND + ELSE
POUNDELSE.ignore( CPPCOMMENT )
POUNDELSE.ignore( CCOMMENT )
POUNDELSE.ignore( FCOMMENT )

#endif
ENDIF = pp.CaselessKeyword('endif')
POUNDENDIF = pp.LineStart() + pp.Group( POUND + ENDIF )
POUNDENDIF.ignore( CPPCOMMENT )
POUNDENDIF.ignore( CCOMMENT )
POUNDENDIF.ignore( FCOMMENT )

#define
DEFINE = pp.Keyword('define')
POUNDDEFINE = pp.LineStart() + POUND + DEFINE
POUNDDEFINE = POUND + DEFINE

#ifdef
IFDEF = pp.CaselessKeyword('ifdef')
POUNDIFDEF = pp.LineStart() + pp.Group( POUND + IFDEF + pp.Group( pp.restOfLine ) )
POUNDIFDEF.ignore( CPPCOMMENT )
POUNDIFDEF.ignore( CCOMMENT )
POUNDIFDEF.ignore( FCOMMENT )

#ifndef
IFNDEF = pp.CaselessKeyword('ifndef')
POUNDIFNDEF = pp.LineStart() + pp.Group( POUND + IFNDEF + pp.Group( pp.restOfLine ) )
POUNDIFNDEF.ignore( CPPCOMMENT )
POUNDIFNDEF.ignore( CCOMMENT )
POUNDIFNDEF.ignore( FCOMMENT )

#undef
UNDEF = pp.CaselessKeyword('undef')
POUNDUNDEF = pp.LineStart() + pp.Group( POUND + UNDEF + pp.Group( pp.restOfLine ) )
POUNDUNDEF.ignore( CPPCOMMENT )
POUNDUNDEF.ignore( CCOMMENT )
POUNDUNDEF.ignore( FCOMMENT )

#
# #define constant value
#
CONSTANTDEFINE = pp.Suppress( POUNDDEFINE ) + gram.VALIDNAME + \
    ( gram.VALIDNAME | gram.HEX_NUM | gram.ENGINEERING_NUM | gram.DECIMAL_NUM | gram.INTEGER | gram.NUM )
CONSTANTDEFINE.ignore( CPPCOMMENT )
CONSTANTDEFINE.ignore( CCOMMENT )
CONSTANTDEFINE.ignore( FCOMMENT )

#
# #if ... #else ... #endif
#
CHUNK = pp.Group( POUNDIF | POUNDIFDEF | POUNDIFNDEF ) + \
    pp.Group( pp.ZeroOrMore( LINES ) ) + pp.Optional(POUNDELSE) + pp.Group( pp.ZeroOrMore( LINES ) ) + \
    pp.Group( POUNDENDIF )
CHUNK.ignore( CPPCOMMENT )
CHUNK.ignore( CCOMMENT )
CHUNK.ignore( FCOMMENT )

CHUNK2 = pp.Group( POUNDIFDEF + pp.Group( pp.ZeroOrMore( LINES ) ) + pp.Optional( POUNDELSE ) + \
                      pp.Group( pp.ZeroOrMore( LINES ) ) + POUNDENDIF )
CHUNK2.ignore( CPPCOMMENT )
CHUNK2.ignore( CCOMMENT )
CHUNK2.ignore( FCOMMENT )

def _parseCONSTANTDEFINE( ln ):
    try:
        for match,start,end in CONSTANTDEFINE.scanString( ln ):
            print match
    except pp.ParseException, pe:
        print pe.msg

if __name__ == "__main__":
    _parseCONSTANTDEFINE('#define apple 4')
    _parseCONSTANTDEFINE('/* just a comment */')
    _parseCONSTANTDEFINE('# define hello 1 ! fcomment')
    _parseCONSTANTDEFINE('#\tdefine hello 1 /* ccomment */')
    _parseCONSTANTDEFINE('#\tdefine hello -3.46e-01')
    _parseCONSTANTDEFINE(' #define cuMemcpyHtoD cuMemcpyHtoD_v2')
    _parseCONSTANTDEFINE('    #define cuDeviceTotalMem                    cuDeviceTotalMem_v2')
    _parseCONSTANTDEFINE('''
#\tdefine hellofloat 3.32e-01 /* comment */
#define hellohex 0x01
#define helloint -42
''')

    tmpstr = """
#ifdef NEVERDEFINE
  do something
  and more
#else /* random comment */
  do other
  do do
#endif

#ifndef NEVERDEFINE
  unsigned int x,y,z;
#endif
"""
    print "POUNDIFDEF ------------------"
    for match,start,end in POUNDIFDEF.scanString( tmpstr ):
        print match
    print "POUNDIFNDEF ------------------"
    for match,start,end in POUNDIFNDEF.scanString( tmpstr ):
        print match
    print "POUNDELSE ------------------"
    for match,start,end in POUNDELSE.scanString( tmpstr ):
        print match
    print "POUNDENDIF ------------------"
    for match,start,end in POUNDENDIF.scanString( tmpstr ):
        print match
    print "CHUNK -----------------------"
    for match,start,end in CHUNK.scanString( tmpstr ):
        print match
    print "CHUNK2 -----------------------"
    for match,start,end in CHUNK2.scanString( tmpstr ):
        print match
    print "LINES -----------------------"
    for match,start,end in LINES.scanString( tmpstr ):
        print match

    tmpstr = """
# 145 "<built-in>"
# 438 "/usr/local/cuda/include/driver_types.h" 1 2 4
  cudaSuccess = 0,
# 452 "/usr/local/cuda/include/driver_types.h"
  cudaErrorReturnFailure,
"""
    print "POUNDNUMLINES ---------------"
    for match,start,end in POUNDNUMFILE.scanString( tmpstr ):
        print match
