#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
from cStringIO import StringIO as SrcBuffer
import copy
import hashlib

class LineScanner(object):
    def __init__(self):
        self._buff = SrcBuffer()
        self.reset()

    def reset(self):
        self._buff.reset()
        self._buff.truncate()
        self._curLine = ''
        self._lineNumber = 0
        self._curByte = 0
        self._lastByte = 0
        self._lineByteMap = []
        self._name = ''

    def __str__(self):
        return self._name

    def _makeMap(self):
        buff = self._buff
        lineByteMap = self._lineByteMap
        buff.seek(0)
        for line in buff:
            lineByteMap.append(buff.tell())
        buff.seek(0)

    def nextLine(self):
        buff = self._buff
        self._lastByte = buff.tell()
        self._curLine = buff.readline()

        if self._curLine == '':
            raise EOFError("EOF")

        self._lineNumber += 1
        self._curByte = buff.tell()
        return self._curLine

    def prevLine(self):
        _n = max(1, self._lineNumber-2)
        self._buff.seek(self._lineByteMap[_n])
        return self.nextLine()

    def backup(self, n=1):
        _lineNum = max(self._lineNumber-n, 1)
        _lineNum = min(_lineNum, len(self._lineByteMap))
        self._lineNumber = _lineNum
        self._buff.seek(self._lineByteMap[_lineNum-1])

    def curLine(self):
        return self._curLine

    def gotoLine(self, n):
        self._buff.seek(self._lineByteMap[n-1])
        self._lineNumber = n
        return self.nextLine()

    def lineNumber(self):
        return self._lineNumber

    def load(self, name):
        '''Load a file into the buffer.

        '''
        self.reset()
        
        try:
            fd = open(name)
            try:
                self._name = name
                self._buff.write(fd.read())
                self._makeMap()
            finally:
                fd.close()
        except:
            print "Could not read %s"%(name,)

    def loadBuff(self, buff, name):
        self.reset()
        self._name = name
        self._buff.write(buff.read())
        self._makeMap()

    def loadString(self, s, name):
        self.reset()
        self._name = name
        self._buff.write(s)
        self._makeMap()

    def hash(self):
        cur = self._buff.tell()
        self._buff.seek(0)
        hsh = hashlib.sha256(self._buff.read()).hexdigest()
        self._buff.seek(cur)
        return hsh

    def name(self):
        return self._name

