#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
__all__ = ['ASet']

import copy

class ASet(set):
    '''A set class that simply has a new "changed" attribute to indicate
    whether it has been updated or added to since last check.

    '''
    def __init__(self, *args, **kwargs):
        super(ASet, self).__init__(*args, **kwargs)
        self._changed = False
        self._changeSet = set()

    def update(self, *args, **kwargs):
        '''Wrapper for the set update() method with changed setting.

        '''
        l = len(self)
        super(ASet, self).update(*args, **kwargs)
        if l != len(self): 
            self._changeSet.update(*args, **kwargs)
            self._changed = True

    def add(self, *args, **kwargs):
        '''Wrapper for the set add() method with changed setting. 

        '''
        l = len(self)
        super(ASet, self).add(*args, **kwargs)
        if l != len(self):
            self._changeSet.add(*args, **kwargs)
            self._changed = True

    def changeSet(self):
        '''Return a copy of only the changes since last reset.

        '''
        return copy.copy(self._changeSet)

    def changed(self):
        return self._changed

    def reset(self):
        '''Reset the changed attribute.

        '''
        self._changed = False
        self._changeSet.clear()

