#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
__all__ = ['SymbolTable']

#................................................................................
# Symbol Table
#................................................................................
class SymbolTable(dict):
    def __init__(self, joinCharacter='|', *args, **kwargs):
        self._joinCharacter = joinCharacter
        super(SymbolTable, self).__init__(*args, **kwargs)

    def joinCharacter(self):
        return self._joinCharacter

    def filter(self, filterFunc):
        _jc = self._joinCharacter

        symbols = [item for item in self.items() if filterFunc(item)]

        table = SymbolTable(_jc)
        table.update(symbols)
        return table

    def copy(self):
        '''Overload the dictionaries copy method.

        '''
        return SymbolTable(self._joinCharacter, self.items())

    def registerAll(self, blk):
        '''Add the symbols in blk to the SymbolTable and all the children as
        well.  The symbol is generated by concatenating child with parent
        name.

        '''
        def _preOrder(blk, symbol):
            self[symbol] = blk
            for name, child in blk.children().items():
                _symbol = _joinCharacter.join([child.name(), symbol])
                _preOrder(child, _symbol)

        _joinCharacter = self._joinCharacter

        if isinstance(blk, dict):
            for b in blk.values():
                _preOrder(b, b.name())
        else:
            _preOrder(blk, blk.name())

    def register(self, blk):
        '''Add a single block to the table. 

        '''
        self[blk.name()] = blk

    def union(self, table):
        '''Merge this table with another symbol table.
        
        '''
        self.update(table)

    def isEmpty(self):
        return (len(self) == 0)

