import unittest
from blockit import *

class TestSymbolTable(unittest.TestCase):
    def setUp(self):
        self.table = SymbolTable()
        prev = Block('test', 'block-0', None)
        prev.setTable(self.table)
        self.blocks = [prev]
        for x in range(1,5):
            typ = 'test'
            name = 'block-%s'%(x,)
            next = Block('test', 'block-%s'%(x,), None)
            prev.setTable(self.table)
            prev.addChild(next)
            prev = next
            self.blocks.append(prev)

    def tearDown(self):
        self.table.clear()

    def testBuild(self):
        '''Test the registering of blocks into the table

        '''
        table = self.table
        blocks = self.blocks

        for blk in blocks:
            table.register(blk)

        for blk in blocks:
            self.assert_(blk.name() in table)
        
        table.registerAll(blocks[0])
        
        sym = blocks[0].name()
        jc = table.joinCharacter()
        for i,blk in enumerate(blocks):
            if i > 0: sym = jc.join([blk.name(),sym])
            self.assert_(sym in table)

        table.clear()
        table.registerAll(blocks[0])
        self.assert_(len(table) == len(blocks))

    def testJoinCharacter(self):
        '''Test a new join character for the table

        '''
        self.table = SymbolTable(':')
        self.testBuild()
        self.table = SymbolTable()

    def testFilter(self):
        '''Test the filter method

        '''
        f1 = lambda (k,v): 'block-1' in k
        f2 = lambda (k,v): v.parent().name() == 'block-0'

        table = self.table
        blocks = self.blocks

        table.registerAll(blocks[0])
        a = table.filter(And(f1, f2))
        self.assert_(len(a) == 1)
        self.assert_('block-1|block-0' in a)

    def testUnion(self):
        '''Test the union of two tables

        '''
        table = self.table
        blocks = self.blocks
        table.registerAll(blocks[0])
        table_ = SymbolTable()
        table_.register(blocks[1])

        table.union(table_)
        self.assert_(len(table) == len(blocks)+1)
        self.assert_(blocks[1].name() in table)
        self.assert_(blocks[0].name() in table)
        table_.clear()

    def testIsEmpty(self):
        '''Test the isEmpty() method

        '''
        self.assert_(self.table.isEmpty())

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSymbolTable)
    unittest.TextTestRunner(verbosity=2).run(suite)

    
